# distutils: libraries = flint
# distutils: depends = flint/fq_nmod.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fq_nmod_ctx_init(fq_nmod_ctx_t ctx, const fmpz_t p, slong d, const char * var) noexcept
    int _fq_nmod_ctx_init_conway(fq_nmod_ctx_t ctx, const fmpz_t p, slong d, const char * var) noexcept
    void fq_nmod_ctx_init_conway(fq_nmod_ctx_t ctx, const fmpz_t p, slong d, const char * var) noexcept
    void fq_nmod_ctx_init_modulus(fq_nmod_ctx_t ctx, const nmod_poly_t modulus, const char * var) noexcept
    void fq_nmod_ctx_clear(fq_nmod_ctx_t ctx) noexcept
    const nmod_poly_struct* fq_nmod_ctx_modulus(const fq_nmod_ctx_t ctx) noexcept
    slong fq_nmod_ctx_degree(const fq_nmod_ctx_t ctx) noexcept
    fmpz * fq_nmod_ctx_prime(const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_ctx_order(fmpz_t f, const fq_nmod_ctx_t ctx) noexcept
    int fq_nmod_ctx_fprint(FILE * file, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_ctx_print(const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_ctx_randtest(fq_nmod_ctx_t ctx, flint_rand_t state) noexcept
    void fq_nmod_ctx_randtest_reducible(fq_nmod_ctx_t ctx, flint_rand_t state) noexcept
    void fq_nmod_init(fq_nmod_t rop, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_init2(fq_nmod_t rop, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_clear(fq_nmod_t rop, const fq_nmod_ctx_t ctx) noexcept
    void _fq_nmod_sparse_reduce(mp_limb_t * R, slong lenR, const fq_nmod_ctx_t ctx) noexcept
    void _fq_nmod_dense_reduce(mp_limb_t * R, slong lenR, const fq_nmod_ctx_t ctx) noexcept
    void _fq_nmod_reduce(mp_limb_t * r, slong lenR, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_reduce(fq_nmod_t rop, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_add(fq_nmod_t rop, const fq_nmod_t op1, const fq_nmod_t op2, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_sub(fq_nmod_t rop, const fq_nmod_t op1, const fq_nmod_t op2, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_sub_one(fq_nmod_t rop, const fq_nmod_t op1, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_neg(fq_nmod_t rop, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_mul(fq_nmod_t rop, const fq_nmod_t op1, const fq_nmod_t op2, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_mul_fmpz(fq_nmod_t rop, const fq_nmod_t op, const fmpz_t x, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_mul_si(fq_nmod_t rop, const fq_nmod_t op, slong x, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_mul_ui(fq_nmod_t rop, const fq_nmod_t op, ulong x, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_sqr(fq_nmod_t rop, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void _fq_nmod_inv(mp_ptr * rop, mp_srcptr * op, slong len, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_inv(fq_nmod_t rop, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_gcdinv(fq_nmod_t f, fq_nmod_t inv, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void _fq_nmod_pow(mp_limb_t * rop, const mp_limb_t * op, slong len, const fmpz_t e, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_pow(fq_nmod_t rop, const fq_nmod_t op, const fmpz_t e, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_pow_ui(fq_nmod_t rop, const fq_nmod_t op, const ulong e, const fq_nmod_ctx_t ctx) noexcept
    int fq_nmod_sqrt(fq_nmod_t rop, const fq_nmod_t op1, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_pth_root(fq_nmod_t rop, const fq_nmod_t op1, const fq_nmod_ctx_t ctx) noexcept
    bint fq_nmod_is_square(const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    int fq_nmod_fprint_pretty(FILE * file, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_print_pretty(const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    int fq_nmod_fprint(FILE * file, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_print(const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    char * fq_nmod_get_str(const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    char * fq_nmod_get_str_pretty(const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_randtest(fq_nmod_t rop, flint_rand_t state, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_randtest_not_zero(fq_nmod_t rop, flint_rand_t state, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_randtest_dense(fq_nmod_t rop, flint_rand_t state, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_rand(fq_nmod_t rop, flint_rand_t state, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_rand_not_zero(fq_nmod_t rop, flint_rand_t state, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_set(fq_nmod_t rop, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_set_si(fq_nmod_t rop, const slong x, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_set_ui(fq_nmod_t rop, const ulong x, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_set_fmpz(fq_nmod_t rop, const fmpz_t x, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_swap(fq_nmod_t op1, fq_nmod_t op2, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_zero(fq_nmod_t rop, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_one(fq_nmod_t rop, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_gen(fq_nmod_t rop, const fq_nmod_ctx_t ctx) noexcept
    int fq_nmod_get_fmpz(fmpz_t rop, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_get_nmod_poly(nmod_poly_t a, const fq_nmod_t b, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_set_nmod_poly(fq_nmod_t a, const nmod_poly_t b, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_get_nmod_mat(nmod_mat_t col, const fq_nmod_t a, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_set_nmod_mat(fq_nmod_t a, const nmod_mat_t col, const fq_nmod_ctx_t ctx) noexcept
    bint fq_nmod_is_zero(const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    bint fq_nmod_is_one(const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    bint fq_nmod_equal(const fq_nmod_t op1, const fq_nmod_t op2, const fq_nmod_ctx_t ctx) noexcept
    bint fq_nmod_is_invertible(const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    bint fq_nmod_is_invertible_f(fq_nmod_t f, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    int fq_nmod_cmp(const fq_nmod_t a, const fq_nmod_t b, const fq_nmod_ctx_t ctx) noexcept
    void _fq_nmod_trace(fmpz_t rop, const mp_limb_t * op, slong len, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_trace(fmpz_t rop, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void _fq_nmod_norm(fmpz_t rop, const mp_limb_t * op, slong len, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_norm(fmpz_t rop, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void _fq_nmod_frobenius(mp_limb_t * rop, const mp_limb_t * op, slong len, slong e, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_frobenius(fq_nmod_t rop, const fq_nmod_t op, slong e, const fq_nmod_ctx_t ctx) noexcept
    int fq_nmod_multiplicative_order(fmpz * ord, const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    bint fq_nmod_is_primitive(const fq_nmod_t op, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_bit_pack(fmpz_t f, const fq_nmod_t op, flint_bitcnt_t bit_size, const fq_nmod_ctx_t ctx) noexcept
    void fq_nmod_bit_unpack(fq_nmod_t rop, const fmpz_t f, flint_bitcnt_t bit_size, const fq_nmod_ctx_t ctx) noexcept



r"""
Alternating forms on free modules

Given a free module `M` of finite rank over a commutative ring `R`
and a positive integer `p`, an *alternating form of degree* `p` on `M` is
a map

.. MATH::

    a:\ \underbrace{M\times\cdots\times M}_{p\ \; \mbox{times}}
    \longrightarrow R

that (i) is multilinear and (ii) vanishes whenever any of two of its
arguments are equal.
An alternating form of degree `p` is a tensor on `M` of type `(0,p)`.

Alternating forms are implemented via the class :class:`FreeModuleAltForm`,
which is a subclass of the generic tensor class
:class:`~sage.tensor.modules.free_module_tensor.FreeModuleTensor`.

AUTHORS:

- Eric Gourgoulhon, Michal Bejger (2014-2015): initial version

REFERENCES:

- Chap. 23 of R. Godement : *Algebra* [God1968]_
- Chap. 15 of S. Lang : *Algebra* [Lan2002]_
"""
# ****************************************************************************
#       Copyright (C) 2015 Eric Gourgoulhon <eric.gourgoulhon@obspm.fr>
#       Copyright (C) 2015 Michal Bejger <bejger@camk.edu.pl>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.tensor.modules.free_module_tensor import FreeModuleTensor
from sage.tensor.modules.comp import Components, CompFullyAntiSym


class FreeModuleAltForm(FreeModuleTensor):
    r"""
    Alternating form on a free module of finite rank over a commutative ring.

    This is a Sage *element* class, the corresponding *parent* class being
    :class:`~sage.tensor.modules.ext_pow_free_module.ExtPowerDualFreeModule`.

    INPUT:

    - ``fmodule`` -- free module `M` of finite rank over a commutative ring
      `R`, as an instance of
      :class:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule`
    - ``degree`` -- positive integer; the degree `p` of the
      alternating form (i.e. its tensor rank)
    - ``name`` -- (default: ``None``) string; name given to the alternating
      form
    - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to denote the
      alternating form; if none is provided, ``name`` is used

    EXAMPLES:

    Alternating form of degree 2 on a rank-3 free module::

        sage: M = FiniteRankFreeModule(ZZ, 3, name='M', start_index=1)
        sage: e = M.basis('e')
        sage: a = M.alternating_form(2, name='a') ; a
        Alternating form a of degree 2 on the
         Rank-3 free module M over the Integer Ring
        sage: type(a)
        <class 'sage.tensor.modules.ext_pow_free_module.ExtPowerDualFreeModule_with_category.element_class'>
        sage: a.parent()
        2nd exterior power of the dual of the Rank-3 free module M over the Integer Ring
        sage: a[1,2], a[2,3] = 4, -3
        sage: a.display(e)
        a = 4 e^1∧e^2 - 3 e^2∧e^3

    The alternating form acting on the basis elements::

        sage: a(e[1],e[2])
        4
        sage: a(e[1],e[3])
        0
        sage: a(e[2],e[3])
        -3
        sage: a(e[2],e[1])
        -4

    An alternating form of degree 1 is a linear form::

        sage: b = M.linear_form('b') ; b
        Linear form b on the Rank-3 free module M over the Integer Ring
        sage: b[:] = [2,-1,3]  # components w.r.t. the module's default basis (e)

    A linear form is a tensor of type `(0,1)`::

        sage: b.tensor_type()
        (0, 1)

    It is an element of the dual module::

        sage: b.parent()
        Dual of the Rank-3 free module M over the Integer Ring
        sage: b.parent() is M.dual()
        True

    The members of a dual basis are linear forms::

        sage: e.dual_basis()[1]
        Linear form e^1 on the Rank-3 free module M over the Integer Ring
        sage: e.dual_basis()[2]
        Linear form e^2 on the Rank-3 free module M over the Integer Ring
        sage: e.dual_basis()[3]
        Linear form e^3 on the Rank-3 free module M over the Integer Ring

    Any linear form is expanded onto them::

        sage: b.display(e)
        b = 2 e^1 - e^2 + 3 e^3

    In the above example, an equivalent writing would have been
    ``b.display()``, since the basis ``e`` is the module's default basis.
    A linear form maps module elements to ring elements::

        sage: v = M([1,1,1])
        sage: b(v)
        4
        sage: b(v) in M.base_ring()
        True

    Test of linearity::

        sage: u = M([-5,-2,7])
        sage: b(3*u - 4*v) == 3*b(u) - 4*b(v)
        True

    The standard tensor operations apply to alternating forms, like the
    extraction of components with respect to a given basis::

        sage: a[e,1,2]
        4
        sage: a[1,2]  # since e is the module's default basis
        4
        sage: all( a[i,j] == - a[j,i] for i in {1,2,3} for j in {1,2,3} )
        True

    the tensor product::

        sage: c = b*b ; c
        Symmetric bilinear form  b⊗b on the Rank-3 free module M over the
         Integer Ring
        sage: c.parent()
        Free module of type-(0,2) tensors on the Rank-3 free module M over the
         Integer Ring
        sage: c.display(e)
        b⊗b = 4 e^1⊗e^1 - 2 e^1⊗e^2 + 6 e^1⊗e^3 - 2 e^2⊗e^1 + e^2⊗e^2
         - 3 e^2⊗e^3 + 6 e^3⊗e^1 - 3 e^3⊗e^2 + 9 e^3⊗e^3

    the contractions::

        sage: s = a.contract(v) ; s
        Linear form on the Rank-3 free module M over the Integer Ring
        sage: s.parent()
        Dual of the Rank-3 free module M over the Integer Ring
        sage: s.display(e)
        4 e^1 - 7 e^2 + 3 e^3

    or tensor arithmetics::

        sage: s = 3*a + c ; s
        Type-(0,2) tensor on the Rank-3 free module M over the Integer Ring
        sage: s.parent()
        Free module of type-(0,2) tensors on the Rank-3 free module M over the
         Integer Ring
        sage: s.display(e)
        4 e^1⊗e^1 + 10 e^1⊗e^2 + 6 e^1⊗e^3 - 14 e^2⊗e^1 + e^2⊗e^2
         - 12 e^2⊗e^3 + 6 e^3⊗e^1 + 6 e^3⊗e^2 + 9 e^3⊗e^3

    Note that tensor arithmetics preserves the alternating character if both
    operands are alternating::

        sage: s = a - 2*a ; s
        Alternating form of degree 2 on the Rank-3 free module M over the
         Integer Ring
        sage: s.parent() # note the difference with s = 3*a + c above
        2nd exterior power of the dual of the Rank-3 free module M over the
         Integer Ring
        sage: s == -a
        True

    An operation specific to alternating forms is of course the exterior
    product::

        sage: s = a.wedge(b) ; s
        Alternating form a∧b of degree 3 on the Rank-3 free module M over the
         Integer Ring
        sage: s.parent()
        3rd exterior power of the dual of the Rank-3 free module M over the
         Integer Ring
        sage: s.display(e)
        a∧b = 6 e^1∧e^2∧e^3
        sage: s[1,2,3] == a[1,2]*b[3] + a[2,3]*b[1] + a[3,1]*b[2]
        True

    The exterior product is nilpotent on linear forms::

        sage: s = b.wedge(b) ; s
        Alternating form zero of degree 2 on the Rank-3 free module M over the
         Integer Ring
        sage: s.display(e)
        zero = 0
    """
    def __init__(self, fmodule, degree, name=None, latex_name=None):
        r"""
        Initialize ``self``.

        TESTS::

            sage: from sage.tensor.modules.free_module_alt_form import FreeModuleAltForm
            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: a = FreeModuleAltForm(M, 2, name='a')
            sage: a[e,0,1] = 2
            sage: TestSuite(a).run(skip='_test_category') # see below

        In the above test suite, _test_category fails because a is not an
        instance of a.parent().category().element_class. Actually alternating
        forms must be constructed via ExtPowerDualFreeModule.element_class and
        not by a direct call to FreeModuleAltForm::

            sage: a1 = M.dual_exterior_power(2).element_class(M, 2, name='a')
            sage: a1[e,0,1] = 2
            sage: TestSuite(a1).run()
        """
        FreeModuleTensor.__init__(self, fmodule, (0, degree), name=name,
                                  latex_name=latex_name,
                                  antisym=range(degree),
                                  parent=fmodule.dual_exterior_power(degree))

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: M.alternating_form(1)
            Linear form on the Rank-3 free module M over the Integer Ring
            sage: M.alternating_form(1, name='a')
            Linear form a on the Rank-3 free module M over the Integer Ring
            sage: M.alternating_form(2)
            Alternating form of degree 2 on the
             Rank-3 free module M over the Integer Ring
            sage: M.alternating_form(2, name='a')
            Alternating form a of degree 2 on the
             Rank-3 free module M over the Integer Ring
        """
        if self._tensor_rank == 1:
            description = "Linear form "
            if self._name is not None:
                description += self._name + " "
        else:
            description = "Alternating form "
            if self._name is not None:
                description += self._name + " "
            description += "of degree {} ".format(self._tensor_rank)
        description += "on the {}".format(self._fmodule)
        return description

    def _new_instance(self):
        r"""
        Create an instance of the same class as ``self``, on the same module
        and of the same degree.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: a = M.alternating_form(1)
            sage: a._new_instance()
            Linear form on the Rank-3 free module M over the Integer Ring
            sage: a._new_instance().parent() is a.parent()
            True
            sage: b = M.alternating_form(2, name='b')
            sage: b._new_instance()
            Alternating form of degree 2 on the
             Rank-3 free module M over the Integer Ring
            sage: b._new_instance().parent() is b.parent()
            True
        """
        return self.__class__(self._fmodule, self._tensor_rank)

    def _new_comp(self, basis):
        r"""
        Create some (uninitialized) components of ``self`` in a given basis.

        This method, which is already implemented in
        :meth:`FreeModuleTensor._new_comp`, is redefined here for efficiency.

        INPUT:

        - ``basis`` -- basis of the free module on which ``self`` is defined

        OUTPUT:

        - an instance of :class:`~sage.tensor.modules.comp.CompFullyAntiSym`,
          or of :class:`~sage.tensor.modules.comp.Components` if
          the degree of ``self`` is 1.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: a = M.alternating_form(2, name='a')
            sage: a._new_comp(e)
            Fully antisymmetric 2-indices components w.r.t. Basis (e_0,e_1,e_2)
             on the Rank-3 free module M over the Integer Ring
            sage: a = M.alternating_form(1)
            sage: a._new_comp(e)
            1-index components w.r.t. Basis (e_0,e_1,e_2) on the Rank-3 free
             module M over the Integer Ring
        """
        fmodule = self._fmodule  # the base free module
        if self._tensor_rank == 1:
            return Components(fmodule._ring, basis, 1,
                              start_index=fmodule._sindex,
                              output_formatter=fmodule._output_formatter)

        return CompFullyAntiSym(fmodule._ring, basis, self._tensor_rank,
                                start_index=fmodule._sindex,
                                output_formatter=fmodule._output_formatter)

    def degree(self):
        r"""
        Return the degree of ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: a = M.alternating_form(2, name='a')
            sage: a.degree()
            2
        """
        return self._tensor_rank

    def _display_expansion(self, basis=None, format_spec=None):
        r"""
        Return the pure expansion of ``self`` w.r.t. a given module basis.

        The expansion is actually performed onto exterior products of elements
        of the cobasis (dual basis) associated with ``basis`` (see examples
        below). The output is either text-formatted (console mode) or
        LaTeX-formatted (notebook mode).

        INPUT:

        - ``basis`` -- (default: ``None``) basis of the free module with
          respect to which the alternating form is expanded; if none is
          provided, the module's default basis is assumed
        - ``format_spec`` -- (default: ``None``) format specification passed
          to ``self._fmodule._output_formatter`` to format the output

        TESTS:

        Expansion display of an alternating form of degree 1 (linear form) on a
        rank-3 free module::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: e.dual_basis()
            Dual basis (e^0,e^1,e^2) on the Rank-3 free module M over the Integer Ring
            sage: a = M.linear_form('a', latex_name=r'\alpha')
            sage: a[:] = [1,-3,4]
            sage: a._display_expansion(e)
            e^0 - 3 e^1 + 4 e^2
            sage: a._display_expansion() # a shortcut since e is M's default basis
            e^0 - 3 e^1 + 4 e^2
            sage: latex(a._display_expansion())  # display in the notebook
            e^{0} -3 e^{1} + 4 e^{2}
        """
        from sage.misc.latex import latex
        from sage.typeset.unicode_characters import unicode_wedge
        from .format_utilities import is_atomic, FormattedExpansion
        basis, format_spec = self._preparse_display(basis=basis,
                                                    format_spec=format_spec)
        cobasis = basis.dual_basis()
        comp = self.comp(basis)
        terms_txt = []
        terms_latex = []
        for ind in comp.non_redundant_index_generator():
            ind_arg = ind + (format_spec,)
            coef = comp[ind_arg]
            # Check whether the coefficient is zero, preferably via
            # the fast method is_trivial_zero():
            if hasattr(coef, 'is_trivial_zero'):
                zero_coef = coef.is_trivial_zero()
            else:
                zero_coef = coef == 0
            if not zero_coef:
                bases_txt = []
                bases_latex = []
                for k in range(self._tensor_rank):
                    bases_txt.append(cobasis[ind[k]]._name)
                    bases_latex.append(latex(cobasis[ind[k]]))
                basis_term_txt = unicode_wedge.join(bases_txt)
                basis_term_latex = r'\wedge '.join(bases_latex)
                coef_txt = repr(coef)
                if coef_txt == '1':
                    terms_txt.append(basis_term_txt)
                    terms_latex.append(basis_term_latex)
                elif coef_txt == '-1':
                    terms_txt.append('-' + basis_term_txt)
                    terms_latex.append('-' + basis_term_latex)
                else:
                    coef_latex = latex(coef)
                    if is_atomic(coef_txt):
                        terms_txt.append(coef_txt + ' ' + basis_term_txt)
                    else:
                        terms_txt.append('(' + coef_txt + ') ' +
                                         basis_term_txt)
                    if is_atomic(coef_latex):
                        terms_latex.append(coef_latex + basis_term_latex)
                    else:
                        terms_latex.append(r'\left(' + coef_latex +
                                           r'\right)' + basis_term_latex)
        if not terms_txt:
            expansion_txt = '0'
        else:
            expansion_txt = terms_txt[0]
            for term in terms_txt[1:]:
                if term[0] == '-':
                    expansion_txt += ' - ' + term[1:]
                else:
                    expansion_txt += ' + ' + term
        if not terms_latex:
            expansion_latex = '0'
        else:
            expansion_latex = terms_latex[0]
            for term in terms_latex[1:]:
                if term[0] == '-':
                    expansion_latex += term
                else:
                    expansion_latex += '+' + term
        return FormattedExpansion(expansion_txt, expansion_latex)

    def display(self, basis=None, format_spec=None):
        r"""
        Display the alternating form ``self`` in terms of its expansion w.r.t.
        a given module basis.

        The expansion is actually performed onto exterior products of elements
        of the cobasis (dual basis) associated with ``basis`` (see examples
        below). The output is either text-formatted (console mode) or
        LaTeX-formatted (notebook mode).

        INPUT:

        - ``basis`` -- (default: ``None``) basis of the free module with
          respect to which the alternating form is expanded; if none is
          provided, the module's default basis is assumed
        - ``format_spec`` -- (default: ``None``) format specification passed
          to ``self._fmodule._output_formatter`` to format the output

        EXAMPLES:

        Display of an alternating form of degree 1 (linear form) on a rank-3
        free module::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: e.dual_basis()
            Dual basis (e^0,e^1,e^2) on the Rank-3 free module M over the Integer Ring
            sage: a = M.linear_form('a', latex_name=r'\alpha')
            sage: a[:] = [1,-3,4]
            sage: a.display(e)
            a = e^0 - 3 e^1 + 4 e^2
            sage: a.display()  # a shortcut since e is M's default basis
            a = e^0 - 3 e^1 + 4 e^2
            sage: latex(a.display())  # display in the notebook
            \alpha = e^{0} -3 e^{1} + 4 e^{2}

        A shortcut is ``disp()``::

            sage: a.disp()
            a = e^0 - 3 e^1 + 4 e^2

        Display of an alternating form of degree 2 on a rank-3 free module::

            sage: b = M.alternating_form(2, 'b', latex_name=r'\beta')
            sage: b[0,1], b[0,2], b[1,2] = 3, 2, -1
            sage: b.display()
            b = 3 e^0∧e^1 + 2 e^0∧e^2 - e^1∧e^2
            sage: latex(b.display())  # display in the notebook
            \beta = 3 e^{0}\wedge e^{1} + 2 e^{0}\wedge e^{2} -e^{1}\wedge e^{2}

        Display of an alternating form of degree 3 on a rank-3 free module::

            sage: c = M.alternating_form(3, 'c')
            sage: c[0,1,2] = 4
            sage: c.display()
            c = 4 e^0∧e^1∧e^2
            sage: latex(c.display())
            c = 4 e^{0}\wedge e^{1}\wedge e^{2}

        Display of a vanishing alternating form::

            sage: c[0,1,2] = 0  # the only independent component set to zero
            sage: c.is_zero()
            True
            sage: c.display()
            c = 0
            sage: latex(c.display())
            c = 0
            sage: c[0,1,2] = 4  # value restored for what follows

        Display in a basis which is not the default one::

            sage: aut = M.automorphism(matrix=[[0,1,0], [0,0,-1], [1,0,0]],
            ....:                      basis=e)
            sage: f = e.new_basis(aut, 'f')
            sage: a.display(f)
            a = 4 f^0 + f^1 + 3 f^2
            sage: a.disp(f)     # shortcut notation
            a = 4 f^0 + f^1 + 3 f^2
            sage: b.display(f)
            b = -2 f^0∧f^1 - f^0∧f^2 - 3 f^1∧f^2
            sage: c.display(f)
            c = -4 f^0∧f^1∧f^2

        The output format can be set via the argument ``output_formatter``
        passed at the module construction::

            sage: N = FiniteRankFreeModule(QQ, 3, name='N', start_index=1,
            ....:                   output_formatter=Rational.numerical_approx)
            sage: e = N.basis('e')
            sage: b = N.alternating_form(2, 'b')
            sage: b[1,2], b[1,3], b[2,3] = 1/3, 5/2, 4
            sage: b.display()  # default format (53 bits of precision)
            b = 0.333333333333333 e^1∧e^2 + 2.50000000000000 e^1∧e^3
             + 4.00000000000000 e^2∧e^3

        The output format is then controlled by the argument ``format_spec`` of
        the method :meth:`display`::

            sage: b.display(format_spec=10)  # 10 bits of precision
            b = 0.33 e^1∧e^2 + 2.5 e^1∧e^3 + 4.0 e^2∧e^3

        Check that the bug reported in :issue:`22520` is fixed::

            sage: # needs sage.symbolic
            sage: M = FiniteRankFreeModule(SR, 2, name='M')
            sage: e = M.basis('e')
            sage: a = M.alternating_form(2)
            sage: a[0,1] = SR.var('t', domain='real')
            sage: a.display()
            t e^0∧e^1
        """
        from sage.misc.latex import latex
        from sage.tensor.modules.format_utilities import FormattedExpansion
        exp = self._display_expansion(basis=basis, format_spec=format_spec)
        if self._name is None:
            resu_txt = repr(exp)
        else:
            resu_txt = self._name + " = " + repr(exp)
        if self._latex_name is None:
            resu_latex = latex(exp)
        else:
            resu_latex = latex(self) + " = " + latex(exp)
        return FormattedExpansion(resu_txt, resu_latex)

    disp = display

    def wedge(self, other):
        r"""
        Exterior product of ``self`` with the alternating form ``other``.

        INPUT:

        - ``other`` -- an alternating form

        OUTPUT:

        - instance of :class:`FreeModuleAltForm` representing the exterior
          product ``self ∧ other``

        EXAMPLES:

        Exterior product of two linear forms::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: a = M.linear_form('A')
            sage: a[:] = [1,-3,4]
            sage: b = M.linear_form('B')
            sage: b[:] = [2,-1,2]
            sage: c = a.wedge(b) ; c
            Alternating form A∧B of degree 2 on the Rank-3 free module M
             over the Integer Ring
            sage: c.display()
            A∧B = 5 e^0∧e^1 - 6 e^0∧e^2 - 2 e^1∧e^2
            sage: latex(c)
            A\wedge B
            sage: latex(c.display())
            A\wedge B = 5 e^{0}\wedge e^{1} -6 e^{0}\wedge e^{2} -2 e^{1}\wedge e^{2}

        Test of the computation::

            sage: a.wedge(b) == a*b - b*a
            True

        Exterior product of a linear form and an alternating form of degree 2::

            sage: d = M.linear_form('D')
            sage: d[:] = [-1,2,4]
            sage: s = d.wedge(c) ; s
            Alternating form D∧A∧B of degree 3 on the Rank-3 free module M
             over the Integer Ring
            sage: s.display()
            D∧A∧B = 34 e^0∧e^1∧e^2

        Test of the computation::

            sage: s[0,1,2] == d[0]*c[1,2] + d[1]*c[2,0] + d[2]*c[0,1]
            True

        Let us check that the exterior product is associative::

            sage: d.wedge(a.wedge(b)) == (d.wedge(a)).wedge(b)
            True

        and that it is graded anticommutative::

            sage: a.wedge(b) == - b.wedge(a)
            True
            sage: d.wedge(c) == c.wedge(d)
            True
        """
        from sage.typeset.unicode_characters import unicode_wedge
        from .format_utilities import is_atomic
        if not isinstance(other, FreeModuleAltForm):
            raise TypeError("the second argument for the exterior product " +
                            "must be an alternating form")
        if other._tensor_rank == 0:
            return other * self
        if self._tensor_rank == 0:
            return self * other
        fmodule = self._fmodule
        rank_r = self._tensor_rank + other._tensor_rank
        # Facilitate computations involving zero:
        if rank_r > fmodule._rank:
            return fmodule.dual_exterior_power(rank_r).zero()
        if self._is_zero or other._is_zero:
            return fmodule.dual_exterior_power(rank_r).zero()
        if self is other and (self._tensor_rank % 2) == 1:
            return fmodule.dual_exterior_power(rank_r).zero()
        # Generic case:
        basis = self.common_basis(other)
        if basis is None:
            raise ValueError("no common basis for the exterior product")
        cmp_s = self._components[basis]
        cmp_o = other._components[basis]
        cmp_r = CompFullyAntiSym(fmodule._ring, basis, rank_r,
                                 start_index=fmodule._sindex,
                                 output_formatter=fmodule._output_formatter)
        for ind_s, val_s in cmp_s._comp.items():
            for ind_o, val_o in cmp_o._comp.items():
                ind_r = ind_s + ind_o
                if len(ind_r) == len(set(ind_r)):  # all indices are different
                    cmp_r[[ind_r]] += val_s * val_o
        result = fmodule.alternating_form(rank_r)
        result._components[basis] = cmp_r
        if self._name is not None and other._name is not None:
            sname = self._name
            oname = other._name
            if not is_atomic(sname):
                sname = '(' + sname + ')'
            if not is_atomic(oname):
                oname = '(' + oname + ')'
            result._name = sname + unicode_wedge + oname
        if self._latex_name is not None and other._latex_name is not None:
            slname = self._latex_name
            olname = other._latex_name
            if not is_atomic(slname):
                slname = '(' + slname + ')'
            if not is_atomic(olname):
                olname = '(' + olname + ')'
            result._latex_name = slname + r'\wedge ' + olname
        return result

    def interior_product(self, alt_tensor):
        r"""
        Interior product with an alternating contravariant tensor.

        If ``self`` is an alternating form `A` of degree `p` and `B` is an
        alternating contravariant tensor of degree `q\geq p` on the same free
        module, the interior product of `A` by `B` is the alternating
        contravariant tensor `\iota_A B` of degree `q-p` defined by

        .. MATH::

            (\iota_A B)^{i_1\ldots i_{q-p}} = A_{k_1\ldots k_p}
                            B^{k_1\ldots k_p i_1\ldots i_{q-p}}

        .. NOTE::

            ``A.interior_product(B)`` yields the same result as
            ``A.contract(0,..., p-1, B, 0,..., p-1)`` (cf.
            :meth:`~sage.tensor.modules.free_module_tensor.FreeModuleTensor.contract`),
            but ``interior_product`` is more efficient, the alternating
            character of `A` being not used to reduce the computation in
            :meth:`~sage.tensor.modules.free_module_tensor.FreeModuleTensor.contract`

        INPUT:

        - ``alt_tensor`` -- alternating contravariant tensor `B` (instance of
          :class:`~sage.tensor.modules.alternating_contr_tensor.AlternatingContrTensor`);
          the degree of `B` must be at least equal to the degree of ``self``

        OUTPUT:

        - element of the base ring (case `p=q`) or
          :class:`~sage.tensor.modules.alternating_contr_tensor.AlternatingContrTensor`
          (case `p<q`) representing the interior product `\iota_A B`, where `A`
          is ``self``

        .. SEEALSO::

            :meth:`~sage.tensor.modules.alternating_contr_tensor.AlternatingContrTensor.interior_product`
            for the interior product of an alternating contravariant tensor by
            an alternating form

        EXAMPLES:

        Let us consider a rank-3 free module::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M', start_index=1)
            sage: e = M.basis('e')

        and various interior products on it, starting with a linear form
        (``p=1``) and a module element (``q=1``)::

            sage: a = M.linear_form(name='A')
            sage: a[:] = [-2, 4, 3]
            sage: b = M([3, 1, 5], basis=e, name='B')
            sage: c = a.interior_product(b); c
            13
            sage: c == a.contract(b)
            True

        Case  ``p=1`` and ``q=2``::

            sage: b = M.alternating_contravariant_tensor(2, name='B')
            sage: b[1,2], b[1,3], b[2,3] = 5, 2, 3
            sage: c = a.interior_product(b); c
            Element i_A B of the Rank-3 free module M over the Integer Ring
            sage: c.display()
            i_A B = -26 e_1 - 19 e_2 + 8 e_3
            sage: latex(c)
            \iota_{A} B
            sage: c == a.contract(b)
            True

        Case  ``p=1`` and ``q=3``::

            sage: b = M.alternating_contravariant_tensor(3, name='B')
            sage: b[1,2,3] = 5
            sage: c = a.interior_product(b); c
            Alternating contravariant tensor i_A B of degree 2 on the Rank-3 free module M over the Integer Ring
            sage: c.display()
            i_A B = 15 e_1∧e_2 - 20 e_1∧e_3 - 10 e_2∧e_3
            sage: c == a.contract(b)
            True

        Case  ``p=2`` and ``q=2``::

            sage: a = M.alternating_form(2, name='A')
            sage: a[1,2], a[1,3], a[2,3] = 2, -3, 1
            sage: b = M.alternating_contravariant_tensor(2, name='B')
            sage: b[1,2], b[1,3], b[2,3] = 5, 2, 3
            sage: c = a.interior_product(b); c
            14
            sage: c == a.contract(0, 1, b, 0, 1)   # contraction on all indices of a
            True

        Case  ``p=2`` and ``q=3``::

            sage: b = M.alternating_contravariant_tensor(3, name='B')
            sage: b[1,2,3] = 5
            sage: c = a.interior_product(b); c
            Element i_A B of the Rank-3 free module M over the Integer Ring
            sage: c.display()
            i_A B = 10 e_1 + 30 e_2 + 20 e_3
            sage: c == a.contract(0, 1, b, 0, 1)
            True

        Case  ``p=3`` and ``q=3``::

            sage: a = M.alternating_form(3, name='A')
            sage: a[1,2,3] = -2
            sage: c = a.interior_product(b); c
            -60
            sage: c  == a.contract(0, 1, 2, b, 0, 1, 2)
            True
        """
        from .format_utilities import is_atomic
        from .alternating_contr_tensor import AlternatingContrTensor
        if not isinstance(alt_tensor, AlternatingContrTensor):
            raise TypeError("{} is not an alternating ".format(alt_tensor) +
                            "contravariant tensor")
        p_res = alt_tensor._tensor_rank - self._tensor_rank  # degree of result
        if self._tensor_rank == 1:
            # Case p = 1:
            res = self.contract(alt_tensor)
            # contract() deals efficiently with antisymmetry for p = 1
        else:
            # Case p > 1:
            if alt_tensor._fmodule != self._fmodule:
                raise ValueError("{} is not defined on ".format(alt_tensor) +
                                 "the same module as the {}".format(self))
            if alt_tensor._tensor_rank < self._tensor_rank:
                raise ValueError("the degree of the {} ".format(alt_tensor) +
                                 "is lower than that of the {}".format(self))
            # Interior product at the component level:
            basis = self.common_basis(alt_tensor)
            if basis is None:
                raise ValueError("no common basis for the interior product")
            comp = self._components[basis].interior_product(
                                                 alt_tensor._components[basis])
            if p_res == 0:
                res = comp  # result is a scalar
            else:
                res = self._fmodule.tensor_from_comp((p_res, 0), comp)
        # Name of the result
        res_name = None
        if self._name is not None and alt_tensor._name is not None:
            sname = self._name
            oname = alt_tensor._name
            if not is_atomic(sname):
                sname = '(' + sname + ')'
            if not is_atomic(oname):
                oname = '(' + oname + ')'
            res_name = 'i_' + sname + ' ' + oname
        res_latex_name = None
        if self._latex_name is not None and alt_tensor._latex_name is not None:
            slname = self._latex_name
            olname = alt_tensor._latex_name
            if not is_atomic(olname):
                olname = r'\left(' + olname + r'\right)'
            res_latex_name = r'\iota_{' + slname + '} ' + olname
        if res_name:
            try:
                # there is no guarantee that the result has set_name
                # and is mutable
                res.set_name(res_name, latex_name=res_latex_name)
            except (AttributeError, TypeError, ValueError):
                pass
        return res



# sage.doctest: needs sage.rings.padics
r"""
Manin map

Represents maps from a set of right coset representatives to a
coefficient module.

This is a basic building block for implementing modular symbols, and
provides basic arithmetic and right action of matrices.

EXAMPLES::

    sage: E = EllipticCurve('11a')
    sage: phi = E.pollack_stevens_modular_symbol()
    sage: phi
    Modular symbol of level 11 with values in Sym^0 Q^2
    sage: phi.values()
    [-1/5, 1, 0]

    sage: from sage.modular.pollack_stevens.manin_map import ManinMap, M2Z
    sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
    sage: D = OverconvergentDistributions(0, 11, 10)
    sage: MR = ManinRelations(11)
    sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
    sage: f = ManinMap(D, MR, data)
    sage: f(M2Z([1,0,0,1]))
    (1 + O(11^2), 2 + O(11))

    sage: S = Symk(0,QQ)
    sage: MR = ManinRelations(37)
    sage: data  = {M2Z([-2,-3,5,7]): S(0), M2Z([1,0,0,1]): S(0), M2Z([-1,-2,3,5]): S(0), M2Z([-1,-4,2,7]): S(1), M2Z([0,-1,1,4]): S(1), M2Z([-3,-1,7,2]): S(-1), M2Z([-2,-3,3,4]): S(0), M2Z([-4,-3,7,5]): S(0), M2Z([-1,-1,4,3]): S(0)}
    sage: f = ManinMap(S,MR,data)
    sage: f(M2Z([2,3,4,5]))
    1
"""

# ****************************************************************************
#       Copyright (C) 2012 Robert Pollack <rpollack@math.bu.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from typing import Self

from sage.matrix.matrix_space import MatrixSpace
from sage.rings.continued_fraction import convergents
from sage.rings.integer_ring import ZZ
from sage.structure.element import coercion_model

from .fund_domain import M2Z, t00, t01, t10, t11
from .sigma0 import Sigma0


def unimod_matrices_to_infty(r, s):
    r"""
    Return a list of matrices whose associated unimodular paths connect `0` to ``r/s``.

    INPUT:

    - ``r``, ``s`` -- rational numbers

    OUTPUT:

    - a list of matrices in `SL_2(\ZZ)`

    EXAMPLES::

        sage: v = sage.modular.pollack_stevens.manin_map.unimod_matrices_to_infty(19,23); v
        [
        [1 0]  [ 0  1]  [1 4]  [-4  5]  [ 5 19]
        [0 1], [-1  1], [1 5], [-5  6], [ 6 23]
        ]
        sage: [a.det() for a in v]
        [1, 1, 1, 1, 1]

        sage: sage.modular.pollack_stevens.manin_map.unimod_matrices_to_infty(11,25)
        [
        [1 0]  [ 0  1]  [1 3]  [-3  4]  [ 4 11]
        [0 1], [-1  2], [2 7], [-7  9], [ 9 25]
        ]


    ALGORITHM:

    This is Manin's continued fraction trick, which gives an expression
    `\{0,r/s\} = \{0,\infty\} + ... + \{a,b\} + ... + \{*,r/s\}`, where each `\{a,b\}` is
    the image of `\{0,\infty\}` under a matrix in `SL_2(\ZZ)`.
    """
    if s == 0:
        return []
    # the function contfrac_q in
    # https://github.com/williamstein/psage/blob/master/psage/modform/rational/modular_symbol_map.pyx
    # is very, very relevant to massively optimizing this.
    L = convergents(r / s)
    # Computes the continued fraction convergents of r/s
    v = [M2Z([1, L[0].numerator(), 0, L[0].denominator()])]
    # Initializes the list of matrices
    for j in range(len(L) - 1):
        a = L[j].numerator()
        c = L[j].denominator()
        b = L[j + 1].numerator()
        d = L[j + 1].denominator()
        v.append(M2Z([(-1) ** (j + 1) * a, b, (-1) ** (j + 1) * c, d]))
        # The matrix connecting two consecutive convergents is added on
    return v


def unimod_matrices_from_infty(r, s):
    r"""
    Return a list of matrices whose associated unimodular paths connect `\infty` to ``r/s``.

    INPUT:

    - ``r``, ``s`` -- rational numbers

    OUTPUT:

    - a list of `SL_2(\ZZ)` matrices

    EXAMPLES::

        sage: v = sage.modular.pollack_stevens.manin_map.unimod_matrices_from_infty(19,23); v
        [
        [ 0  1]  [-1  0]  [-4  1]  [-5 -4]  [-19   5]
        [-1  0], [-1 -1], [-5  1], [-6 -5], [-23   6]
        ]
        sage: [a.det() for a in v]
        [1, 1, 1, 1, 1]

        sage: sage.modular.pollack_stevens.manin_map.unimod_matrices_from_infty(11,25)
        [
        [ 0  1]  [-1  0]  [-3  1]  [-4 -3]  [-11   4]
        [-1  0], [-2 -1], [-7  2], [-9 -7], [-25   9]
        ]


    ALGORITHM:

    This is Manin's continued fraction trick, which gives an expression
    `\{\infty,r/s\} = \{\infty,0\} + ... + \{a,b\} + ... + \{*,r/s\}`, where each
    `\{a,b\}` is the image of `\{0,\infty\}` under a matrix in `SL_2(\ZZ)`.
    """
    if s != 0:
        L = convergents(r / s)
        # Computes the continued fraction convergents of r/s
        v = [M2Z([-L[0].numerator(), 1, -L[0].denominator(), 0])]
        # Initializes the list of matrices
        # the function contfrac_q in https://github.com/williamstein/psage/blob/master/psage/modform/rational/modular_symbol_map.pyx
        # is very, very relevant to massively optimizing this.
        for j in range(len(L) - 1):
            a = L[j].numerator()
            c = L[j].denominator()
            b = L[j + 1].numerator()
            d = L[j + 1].denominator()
            v.append(M2Z([-b, (-1) ** (j + 1) * a, -d, (-1) ** (j + 1) * c]))
            # The matrix connecting two consecutive convergents is added on
        return v
    return []


class ManinMap:
    r"""
    Map from a set of right coset representatives of `\Gamma_0(N)` in
    `SL_2(\ZZ)` to a coefficient module that satisfies the Manin
    relations.

    INPUT:

    - ``codomain`` -- coefficient module
    - ``manin_relations`` -- a :class:`sage.modular.pollack_stevens.fund_domain.ManinRelations`
      object
    - ``defining_data`` -- dictionary whose keys are a superset of
      ``manin_relations.gens()`` and a subset of ``manin_relations.reps()``,
      and whose values are in the codomain
    - ``check`` -- do numerous (slow) checks and transformations to
      ensure that the input data is perfect

    EXAMPLES::

        sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
        sage: D = OverconvergentDistributions(0, 11, 10)
        sage: manin = sage.modular.pollack_stevens.fund_domain.ManinRelations(11)
        sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
        sage: f = ManinMap(D, manin, data); f # indirect doctest
        Map from the set of right cosets of Gamma0(11) in SL_2(Z) to Space of 11-adic distributions with k=0 action and precision cap 10
        sage: f(M2Z([1,0,0,1]))
        (1 + O(11^2), 2 + O(11))
    """
    def __init__(self, codomain, manin_relations, defining_data, check=True):
        """
        INPUT:

        - ``codomain`` -- coefficient module
        - ``manin_relations`` -- a :class:`ManinRelations` object
        - ``defining_data`` -- dictionary whose keys are a superset of
          :meth:`manin_relations.gens()` and a subset of manin_relations.reps(),
          and whose values are in the codomain
        - ``check`` -- do numerous (slow) checks and transformations to
          ensure that the input data is perfect

        TESTS:

        Test that it fails gracefully on some bogus inputs::

            sage: from sage.modular.pollack_stevens.manin_map import  ManinMap
            sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
            sage: rels = ManinRelations(37)
            sage: ManinMap(ZZ, rels, {})
            Traceback (most recent call last):
            ...
            ValueError: Codomain must have an action of Sigma0(N)
            sage: ManinMap(Symk(0), rels, [])
            Traceback (most recent call last):
            ...
            ValueError: length of defining data must be the same as number of Manin generators
        """
        self._codomain = codomain
        self._manin = manin_relations
        if check:
            if coercion_model.get_action(codomain, Sigma0(manin_relations._N)) is None:
                raise ValueError("Codomain must have an action of Sigma0(N)")
            self._dict = {}
            if isinstance(defining_data, (list, tuple)):
                if len(defining_data) != manin_relations.ngens():
                    raise ValueError("length of defining data must be the same as number of Manin generators")
                for i in range(len(defining_data)):
                    self._dict[manin_relations.gen(i)] = codomain(defining_data[i])
            elif isinstance(defining_data, dict):
                for g in manin_relations.gens():
                    self._dict[g] = codomain(defining_data[g])
            else:
                # constant function
                try:
                    c = codomain(defining_data)
                except TypeError:
                    raise TypeError("unrecognized type %s for defining_data" % type(defining_data))
                g = manin_relations.gens()
                self._dict = dict(zip(g, [c] * len(g)))
        else:
            self._dict = defining_data

    def extend_codomain(self, new_codomain, check=True):
        r"""
        Extend the codomain of ``self`` to ``new_codomain``. There must be a
        valid conversion operation from the old to the new codomain. This is
        most often used for extension of scalars from `\QQ` to `\QQ_p`.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import ManinMap, M2Z
            sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
            sage: S = Symk(0,QQ)
            sage: MR = ManinRelations(37)
            sage: data  = {M2Z([-2,-3,5,7]): S(0), M2Z([1,0,0,1]): S(0), M2Z([-1,-2,3,5]): S(0), M2Z([-1,-4,2,7]): S(1), M2Z([0,-1,1,4]): S(1), M2Z([-3,-1,7,2]): S(-1), M2Z([-2,-3,3,4]): S(0), M2Z([-4,-3,7,5]): S(0), M2Z([-1,-1,4,3]): S(0)}
            sage: m = ManinMap(S, MR, data); m
            Map from the set of right cosets of Gamma0(37) in SL_2(Z) to Sym^0 Q^2
            sage: m.extend_codomain(Symk(0, Qp(11)))
            Map from the set of right cosets of Gamma0(37) in SL_2(Z) to Sym^0 Q_11^2
        """
        new_dict = {}
        for g in self._manin.gens():
            new_dict[g] = new_codomain(self._dict[g])
        return ManinMap(new_codomain, self._manin, new_dict, check)

    def _compute_image_from_gens(self, B):
        r"""
        Compute the image of ``B`` under ``self``.

        INPUT:

        - ``B`` -- generator of Manin relations

        OUTPUT:

        An element in the codomain of ``self`` (e.g. a distribution), the image
        of ``B`` under ``self``.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
            sage: D = OverconvergentDistributions(0, 11, 10)
            sage: MR = ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, MR, data)
            sage: f._compute_image_from_gens(MR.reps()[1])
            (10 + 10*11 + O(11^2), 8 + O(11))
        """
        L = self._manin.relations(B)
        # could raise KeyError if B is not a generator
        t = self._codomain(0)
        for c, A, g in L:
            g1 = self._dict[self._manin.reps(g)] * A
            t += g1 * c
        return t.normalize()

    def __getitem__(self, B):
        r"""

        Compute the image of ``B`` under ``self``.

        INPUT:

        - ``B`` -- coset representative of Manin relations

        OUTPUT:

        An element in the codomain of ``self`` (e.g. a distribution), the image
        of ``B`` under ``self``.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
            sage: S = Symk(0,QQ)
            sage: MR = ManinRelations(37); MR.gens()
            (
            [1 0]  [ 0 -1]  [-1 -1]  [-1 -2]  [-2 -3]  [-3 -1]  [-1 -4]  [-4 -3]
            [0 1], [ 1  4], [ 4  3], [ 3  5], [ 5  7], [ 7  2], [ 2  7], [ 7  5],
            <BLANKLINE>
            [-2 -3]
            [ 3  4]
            )

            sage: data  = {M2Z([-2,-3,5,7]): S(0), M2Z([1,0,0,1]): S(0), M2Z([-1,-2,3,5]): S(0), M2Z([-1,-4,2,7]): S(1), M2Z([0,-1,1,4]): S(1), M2Z([-3,-1,7,2]): S(-1), M2Z([-2,-3,3,4]): S(0), M2Z([-4,-3,7,5]): S(0), M2Z([-1,-1,4,3]): S(0)}
            sage: D = OverconvergentDistributions(2, 37, 40)
            sage: f = ManinMap(D, MR, data)
            sage: f.__getitem__(MR.gens()[1])
            1 + O(37)
            sage: f.__getitem__(MR.gens()[3])
            O(37^40)
            sage: f.__getitem__(MR.gens()[5])
            36 + O(37)
            sage: f[MR.gens()[5]]
            36 + O(37)
        """
        try:
            return self._dict[B]
        except KeyError:
            # To prevent memory overflow
            return self._compute_image_from_gens(B)
            # self._dict[B] = self._compute_image_from_gens(B)
            # return self._dict[B]

    def compute_full_data(self):
        r"""
        Compute the values of ``self`` on all coset reps from its values on our
        generating set.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
            sage: S = Symk(0,QQ)
            sage: MR = ManinRelations(37); MR.gens()
            (
            [1 0]  [ 0 -1]  [-1 -1]  [-1 -2]  [-2 -3]  [-3 -1]  [-1 -4]  [-4 -3]
            [0 1], [ 1  4], [ 4  3], [ 3  5], [ 5  7], [ 7  2], [ 2  7], [ 7  5],
            <BLANKLINE>
            [-2 -3]
            [ 3  4]
            )

            sage: data  = {M2Z([-2,-3,5,7]): S(0), M2Z([1,0,0,1]): S(0), M2Z([-1,-2,3,5]): S(0), M2Z([-1,-4,2,7]): S(1), M2Z([0,-1,1,4]): S(1), M2Z([-3,-1,7,2]): S(-1), M2Z([-2,-3,3,4]): S(0), M2Z([-4,-3,7,5]): S(0), M2Z([-1,-1,4,3]): S(0)}
            sage: f = ManinMap(S,MR,data)
            sage: len(f._dict)
            9
            sage: f.compute_full_data()
            sage: len(f._dict)
            38
        """
        for B in self._manin.reps():
            if B not in self._dict:
                self._dict[B] = self._compute_image_from_gens(B)

    def __add__(self, right):
        r"""
        Return sum ``self + right``, where ``self`` and ``right`` are
        assumed to have identical codomains and Manin relations.

        INPUT:

        - ``self``, ``right`` -- two Manin maps with the same codomain and
          Manin relations

        OUTPUT: the sum of ``self`` and ``right`` -- a Manin map

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: D = OverconvergentDistributions(0, 11, 10); D
            Space of 11-adic distributions with k=0 action and precision cap 10
            sage: manin = sage.modular.pollack_stevens.fund_domain.ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, manin, data); f
            Map from the set of right cosets of Gamma0(11) in SL_2(Z) to Space of 11-adic distributions with k=0 action and precision cap 10
            sage: f(M2Z([1,0,0,1]))
            (1 + O(11^2), 2 + O(11))
            sage: f+f # indirect doctest
            Map from the set of right cosets of Gamma0(11) in SL_2(Z) to Space of 11-adic distributions with k=0 action and precision cap 10
            sage: (f+f)(M2Z([1,0,0,1]))
            (2 + O(11^2), 4 + O(11))
        """
        D = {}
        sd = self._dict
        rd = right._dict
        for ky, val in sd.items():
            if ky in rd:
                D[ky] = val + rd[ky]
        return self.__class__(self._codomain, self._manin, D, check=False)

    def __sub__(self, right):
        """
        Return difference ``self`` - right, where ``self`` and ``right`` are
        assumed to have identical codomains and Manin relations.

        INPUT:

        - ``self``, ``right`` -- two Manin maps with the same codomain and
          Manin relations

        OUTPUT: the difference of ``self`` and ``right`` -- a Manin map

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: D = OverconvergentDistributions(0, 11, 10); D
            Space of 11-adic distributions with k=0 action and precision cap 10
            sage: manin = sage.modular.pollack_stevens.fund_domain.ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, manin, data); f
            Map from the set of right cosets of Gamma0(11) in SL_2(Z) to Space of 11-adic distributions with k=0 action and precision cap 10
            sage: f(M2Z([1,0,0,1]))
            (1 + O(11^2), 2 + O(11))
            sage: f-f
            Map from the set of right cosets of Gamma0(11) in SL_2(Z) to Space of 11-adic distributions with k=0 action and precision cap 10
            sage: (f-f)(M2Z([1,0,0,1]))
            (O(11^2), O(11))
        """
        D = {}
        sd = self._dict
        rd = right._dict
        for ky, val in sd.items():
            if ky in rd:
                D[ky] = val - rd[ky]
        return self.__class__(self._codomain, self._manin, D, check=False)

    def __mul__(self, right):
        """
        Return scalar multiplication ``self * right``, where ``right`` is in
        the base ring of the codomain.

        INPUT:

        - ``self`` -- a Manin map
        - ``right`` -- an element of the base ring of the codomain of self

        OUTPUT: the sum ``self`` and ``right`` -- a Manin map

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: D = OverconvergentDistributions(0, 11, 10)
            sage: manin = sage.modular.pollack_stevens.fund_domain.ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, manin, data)
            sage: f(M2Z([1,0,0,1]))
            (1 + O(11^2), 2 + O(11))
            sage: f*2
            Map from the set of right cosets of Gamma0(11) in SL_2(Z) to Space of 11-adic distributions with k=0 action and precision cap 10
            sage: (f*2)(M2Z([1,0,0,1]))
            (2 + O(11^2), 4 + O(11))
        """
        tp = Sigma0(self._manin.level())(MatrixSpace(ZZ, 2, 2)([1, 0, 0, 1]))
        if isinstance(right, type(tp)):
            return self._right_action(right)

        D = {}
        for ky, val in self._dict.items():
            D[ky] = val * right
        return self.__class__(self._codomain, self._manin, D, check=False)

    def __repr__(self):
        """
        Return string representation of ``self``.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: D = OverconvergentDistributions(0, 11, 10)
            sage: manin = sage.modular.pollack_stevens.fund_domain.ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, manin, data)
            sage: f.__repr__()
            'Map from the set of right cosets of Gamma0(11) in SL_2(Z) to Space of 11-adic distributions with k=0 action and precision cap 10'
        """
        return "Map from the set of right cosets of Gamma0(%s) in SL_2(Z) to %s" % (self._manin.level(), self._codomain)

    def _eval_sl2(self, A):
        r"""
        Return the value of ``self`` on the unimodular divisor corresponding to `A`.

        Note that `A` must be in `SL_2(Z)` for this to work.

        INPUT:

        - ``A`` -- an element of `SL_2(Z)`

        OUTPUT:

        The value of ``self`` on the divisor corresponding to `A` -- i.e. on
        the divisor `\{A(0)\} - \{A(\infty)\}`.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
            sage: D = OverconvergentDistributions(0, 11, 10)
            sage: MR = ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, MR, data)
            sage: A = MR.reps()[1]
            sage: f._eval_sl2(A)
            (10 + 10*11 + O(11^2), 8 + O(11))
        """
        SN = Sigma0(self._manin._N)
        A = M2Z(A)
        B = self._manin.equivalent_rep(A)
        gaminv = SN(B * M2Z(A).adjugate())
        return (self[B] * gaminv).normalize()

    def __call__(self, A):
        """
        Evaluate ``self`` at A.

        INPUT:

        - ``A`` -- a `2 \times 2` matrix

        OUTPUT: the value of ``self`` on the divisor corresponding to ``A`` --
        an element of the codomain of self

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
            sage: D = OverconvergentDistributions(0, 11, 10); D
            Space of 11-adic distributions with k=0 action and precision cap 10
            sage: manin = ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, manin, data); f
            Map from the set of right cosets of Gamma0(11) in SL_2(Z) to Space of 11-adic distributions with k=0 action and precision cap 10
            sage: f(M2Z([1,0,0,1]))
            (1 + O(11^2), 2 + O(11))

            sage: S = Symk(0,QQ)
            sage: MR = ManinRelations(37)
            sage: data  = {M2Z([-2,-3,5,7]): S(0), M2Z([1,0,0,1]): S(0), M2Z([-1,-2,3,5]): S(0), M2Z([-1,-4,2,7]): S(1), M2Z([0,-1,1,4]): S(1), M2Z([-3,-1,7,2]): S(-1), M2Z([-2,-3,3,4]): S(0), M2Z([-4,-3,7,5]): S(0), M2Z([-1,-1,4,3]): S(0)}
            sage: f = ManinMap(S,MR,data)
            sage: f(M2Z([2,3,4,5]))
            1
        """
        a = A[t00]
        b = A[t01]
        c = A[t10]
        d = A[t11]
        # v1: a list of unimodular matrices whose divisors add up to {b/d} - {infty}
        v1 = unimod_matrices_to_infty(b, d)
        # v2: a list of unimodular matrices whose divisors add up to {a/c} - {infty}
        v2 = unimod_matrices_to_infty(a, c)
        # ans: the value of self on A
        ans = self._codomain(0)
        # This loop computes self({b/d}-{infty}) by adding up the values of self on elements of v1
        for B in v1:
            ans = ans + self._eval_sl2(B)

        # This loops subtracts away the value self({a/c}-{infty}) from ans by subtracting away the values of self on elements of v2
        # and so in the end ans becomes self({b/d}-{a/c}) = self({A(0)} - {A(infty)}
        for B in v2:
            ans = ans - self._eval_sl2(B)
        return ans.normalize()

    def apply(self, f, codomain=None, to_moments=False):
        r"""
        Return Manin map given by `x \mapsto f(self(x))`, where `f` is
        anything that can be called with elements of the coefficient
        module.

        This might be used to normalize, reduce modulo a prime, change
        base ring, etc.

        INPUT:

        - ``f`` -- anything that can be called with elements of the coefficient
          module
        - ``codomain`` -- (default: ``None``) the codomain of the return map
        - ``to_moments`` -- boolean (default: ``False``); if ``True``, will
          apply ``f`` to each of the moments instead

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
            sage: S = Symk(0,QQ)
            sage: MR = ManinRelations(37)
            sage: data  = {M2Z([-2,-3,5,7]): S(0), M2Z([1,0,0,1]): S(0), M2Z([-1,-2,3,5]): S(0), M2Z([-1,-4,2,7]): S(1), M2Z([0,-1,1,4]): S(1), M2Z([-3,-1,7,2]): S(-1), M2Z([-2,-3,3,4]): S(0), M2Z([-4,-3,7,5]): S(0), M2Z([-1,-1,4,3]): S(0)}
            sage: f = ManinMap(S,MR,data)
            sage: list(f.apply(lambda t:2*t))
            [0, 2, 0, 0, 0, -2, 2, 0, 0]
        """
        D = {}
        sd = self._dict
        if codomain is None:
            codomain = self._codomain
        for ky, val in sd.items():
            if to_moments:
                D[ky] = codomain([f(val.moment(a))
                                  for a in range(val.precision_absolute())])
            else:
                D[ky] = f(val)
        return self.__class__(codomain, self._manin, D, check=False)

    def __iter__(self):
        r"""
        Return iterator over the values of this map on the reduced
        representatives.

        This might be used to compute the valuation.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: from sage.modular.pollack_stevens.fund_domain import ManinRelations
            sage: S = Symk(0,QQ)
            sage: MR = ManinRelations(37)
            sage: data  = {M2Z([-2,-3,5,7]): S(0), M2Z([1,0,0,1]): S(0), M2Z([-1,-2,3,5]): S(0), M2Z([-1,-4,2,7]): S(1), M2Z([0,-1,1,4]): S(1), M2Z([-3,-1,7,2]): S(-1), M2Z([-2,-3,3,4]): S(0), M2Z([-4,-3,7,5]): S(0), M2Z([-1,-1,4,3]): S(0)}
            sage: f = ManinMap(S,MR,data)
            sage: [a for a in f]
            [0, 1, 0, 0, 0, -1, 1, 0, 0]
        """
        for A in self._manin.gens():
            yield self._dict[A]

    def _right_action(self, gamma):
        r"""
        Return `self | \gamma`, where `\gamma` is a `2 \times 2` integer matrix.

        The action is defined by `(self | \gamma)(D) = self(\gamma D)|\gamma`

        For the action by a single element `\gamma` to be a modular symbol, `\gamma`
        must normalize `\Gamma_0(N)`.  However, this right action
        can also be used to define Hecke operators, in which case each
        individual `self | \gamma` is not a modular symbol on `\Gamma_0(N)`, but
        the sum over acting by the appropriate double coset representatives is.

        INPUT:

        - ``gamma`` -- `2 \times 2` integer matrix of nonzero determinant, with a
          well-defined action on the coefficient module

        OUTPUT: the image of ``self`` under the action of `\gamma` -- a Manin map

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import ManinMap, M2Z, Sigma0
            sage: from sage.modular.pollack_stevens.space import ps_modsym_from_simple_modsym_space
            sage: S01 = Sigma0(1)
            sage: f = Newforms(7, 4)[0]
            sage: f.modular_symbols(1)
            Modular Symbols subspace of dimension 1 of Modular Symbols space of dimension 3 for Gamma_0(7) of weight 4 with sign 1 over Rational Field
            sage: phi = ps_modsym_from_simple_modsym_space(f.modular_symbols(1))._map
            sage: psi = phi._right_action(S01([2,3,4,5])); psi
            Map from the set of right cosets of Gamma0(7) in SL_2(Z) to Sym^2 Q^2

            sage: from sage.modular.pollack_stevens.space import ps_modsym_from_simple_modsym_space
            sage: M = ModularSymbols(17,4,1).cuspidal_subspace()
            sage: A = M.decomposition()
            sage: f = ps_modsym_from_simple_modsym_space(A[0])._map
            sage: g = f._right_action(S01([1,2,0,1]))
            sage: g
            Map from the set of right cosets of Gamma0(17) in SL_2(Z) to Sym^2 Q^2
            sage: x = sage.modular.pollack_stevens.fund_domain.M2Z([2,3,1,0])
            sage: g(x)
            (17, -34, 69)
        """
        D = {}
        # we should eventually replace the for loop with a call to apply_many
        for ky in self._dict:
            D[ky] = self(gamma * ky) * gamma
        return self.__class__(self._codomain, self._manin, D, check=False)

    def normalize(self) -> Self:
        r"""
        Normalize every value of ``self`` -- e.g., reduce each value's
        `j`-th moment modulo `p^{N-j}`.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: D = OverconvergentDistributions(0, 11, 10)
            sage: manin = sage.modular.pollack_stevens.fund_domain.ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, manin, data)
            sage: f._dict[M2Z([1,0,0,1])]
            (1 + O(11^2), 2 + O(11))
            sage: g = f.normalize()
            sage: g._dict[M2Z([1,0,0,1])]
            (1 + O(11^2), 2 + O(11))
        """
        sd = self._dict
        for val in sd.values():
            val.normalize()
        return self

    def reduce_precision(self, M):
        r"""
        Reduce the precision of all the values of the Manin map.

        INPUT:

        - ``M`` -- integer; the new precision

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: D = OverconvergentDistributions(0, 11, 10)
            sage: manin = sage.modular.pollack_stevens.fund_domain.ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, manin, data)
            sage: f._dict[M2Z([1,0,0,1])]
            (1 + O(11^2), 2 + O(11))
            sage: g = f.reduce_precision(1)
            sage: g._dict[M2Z([1,0,0,1])]
            1 + O(11^2)
        """
        D = {}
        for ky, val in self._dict.items():
            D[ky] = val.reduce_precision(M)
        return self.__class__(self._codomain, self._manin, D, check=False)

    def specialize(self, *args):
        r"""
        Specialize all the values of the Manin map to a new coefficient
        module. Assumes that the codomain has a ``specialize`` method, and
        passes all its arguments to that method.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.manin_map import M2Z, ManinMap
            sage: D = OverconvergentDistributions(0, 11, 10)
            sage: manin = sage.modular.pollack_stevens.fund_domain.ManinRelations(11)
            sage: data  = {M2Z([1,0,0,1]):D([1,2]), M2Z([0,-1,1,3]):D([3,5]), M2Z([-1,-1,3,2]):D([1,1])}
            sage: f = ManinMap(D, manin, data)
            sage: g = f.specialize()
            sage: g._codomain
            Sym^0 Z_11^2
        """
        D = {}
        for ky, val in self._dict.items():
            D[ky] = val.specialize(*args)
        return self.__class__(self._codomain.specialize(*args), self._manin,
                              D, check=False)

    def hecke(self, ell, algorithm='prep'):
        r"""
        Return the image of this Manin map under the Hecke operator `T_{\ell}`.

        INPUT:

        - ``ell`` -- a prime

        - ``algorithm`` -- string; either ``'prep'`` (default) or ``'naive'``

        OUTPUT: the image of this ManinMap under the Hecke operator `T_{\ell}`

        EXAMPLES::

            sage: E = EllipticCurve('11a')
            sage: phi = E.pollack_stevens_modular_symbol()
            sage: phi.values()
            [-1/5, 1, 0]
            sage: phi.is_Tq_eigensymbol(7,7,10)
            True
            sage: phi.hecke(7).values()
            [2/5, -2, 0]
            sage: phi.Tq_eigenvalue(7,7,10)
            -2
        """
        self.compute_full_data()
        self.normalize()
        M = self._manin

        if algorithm == 'prep':
            # psi will denote self | T_ell
            psi = {}
            for g in M.gens():
                psi_g = sum((self[h] * A for h, A in M.prep_hecke_on_gen_list(ell, g)), self._codomain(0))
                psi_g.normalize()
                psi[g] = psi_g
            return self.__class__(self._codomain, self._manin,
                                  psi, check=False).normalize()
        if algorithm == 'naive':
            S0N = Sigma0(self._manin.level())
            psi = self._right_action(S0N([1, 0, 0, ell]))
            for a in range(1, ell):
                psi += self._right_action(S0N([1, a, 0, ell]))
            if self._manin.level() % ell != 0:
                psi += self._right_action(S0N([ell, 0, 0, 1]))
            return psi.normalize()
        raise ValueError('Algorithm must be either "naive" or "prep"')

    def p_stabilize(self, p, alpha, V):
        r"""
        Return the `p`-stabilization of ``self`` to level `N*p` on which
        `U_p` acts by `\alpha`.

        INPUT:

        - ``p`` -- a prime

        - ``alpha`` -- a `U_p`-eigenvalue

        - ``V`` -- a space of modular symbols

        OUTPUT: the image of this ManinMap under the Hecke operator `T_{\ell}`

        EXAMPLES::

            sage: E = EllipticCurve('11a')
            sage: phi = E.pollack_stevens_modular_symbol()
            sage: f = phi._map
            sage: V = phi.parent()
            sage: f.p_stabilize(5,1,V)
            Map from the set of right cosets of Gamma0(11) in SL_2(Z) to Sym^0 Q^2
        """
        manin = V.source()
        S0 = Sigma0(self._codomain._act._Np)
        pmat = S0([p, 0, 0, 1])
        D = {}
        scalar = 1 / alpha
        W = self._codomain.change_ring(scalar.parent())
        for g in map(M2Z, manin.gens()):
            # we use scale here so that we do not need to define a
            # construction functor in order to scale by something
            # outside the base ring.
            D[g] = W(self._eval_sl2(g) - (self(pmat * g) * pmat).scale(scalar))
        return self.__class__(W, manin, D, check=False)



# sage.doctest: optional - scipy
r"""
Helper functions for plotting the geometric representation of matroids


AUTHORS:

- Jayant Apte (2014-06-06): initial version

.. NOTE::

    This file provides functions that are called by ``show()`` and ``plot()``
    methods of abstract matroids class. The basic idea is to first decide
    the placement of points in $\mathbb{R}^2$ and then draw lines in geometric
    representation through these points. Point placement procedures such as
    ``addtripts``, ``addnontripts`` together produce ``(x,y)`` tuples
    corresponding to ground set of the matroid in a dictionary.
    These methods provide simple but rigid point placement algorithm.
    Alternatively, one can build the point placement dictionary manually or
    via an optimization that gives aesthetically pleasing point placement (in
    some sense. This is not yet implemented).    One can then use
    ``createline`` function to produce sequence of ``100`` points on a smooth
    curve containing the points in the specified line which in turn uses
    :func:`scipy:scipy.interpolate.splprep` and
    :func:`scipy:scipy.interpolate.splev`.  Then one
    can use sage's graphics primitives ``line``, ``point``, ``text`` and
    ``points`` to produce graphics object containing points (ground set
    elements) and lines (for a rank 3 matroid, these are flats of rank 2 of
    size greater than equal to 3) of the geometric representation of the
    matroid. Loops and parallel elements are added as per conventions in
    [Oxl2011]_ using function ``addlp``. The priority order for point placement
    methods used inside plot() and show() is as follows:

    1. User Specified points dictionary and lineorders
    2. cached point placement dictionary and line orders (a list of ordered
       lists) in M._cached_info (a dictionary)
    3. Internal point placement and orders deciding heuristics
       If a custom point placement and/or line orders is desired, then user
       can simply specify the custom points dictionary as::

            M.cached info = {'plot_positions':<dictionary_of _points>,
                             'plot_lineorders':<list of lists>}


REFERENCES
==========

- [Oxl2011]_ James Oxley, "Matroid Theory, Second Edition". Oxford University
  Press, 2011.

EXAMPLES::

    sage: from sage.matroids import matroids_plot_helpers
    sage: M1 = Matroid(ring=GF(2), matrix=[[1, 0, 0, 0, 1, 1, 1,0,1,0,1],
    ....:                                  [0, 1, 0, 1, 0, 1, 1,0,0,1,0],
    ....:                                  [0, 0, 1, 1, 1, 0, 1,0,0,0,0]])
    sage: pos_dict = {0: (0, 0),  1: (2, 0),  2: (1, 2),  3: (1.5, 1.0),
    ....:             4: (0.5, 1.0),  5: (1.0, 0.0), 6: (1.0, 0.666666666666667),
    ....:             7: (3,3), 8: (4,0), 9: (-1,1), 10: (-2,-2)}
    sage: M1._cached_info = {'plot_positions': pos_dict, 'plot_lineorders': None}
    sage: matroids_plot_helpers.geomrep(M1, sp=True)                                    # needs sage.plot sage.rings.finite_rings
    Graphics object consisting of 22 graphics primitives
"""
# *****************************************************************************
#       Copyright (C) 2013 Jayant Apte <jayant91089@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************

import scipy
import scipy.interpolate
import numpy as np

from sage.matroids.advanced import newlabel
from sage.misc.lazy_import import lazy_import
from sage.sets.set import Set

lazy_import("sage.plot.all", ["Graphics", "line", "text",
                              "polygon2d", "point", "points"])
lazy_import("sage.plot.colors", "Color")


def it(M, B1, nB1, lps) -> tuple[dict, list, list, list]:
    r"""
    Return points on and off the triangle and lines to be drawn for a rank 3
    matroid.

    INPUT:

    - ``M`` -- matroid
    - ``B1`` -- list of groundset elements of ``M`` that corresponds to a
      basis of matroid ``M``
    - ``nB1`` -- list of elements in the ground set of M that corresponds to
      ``M.simplify.groundset() \ B1``
    - ``lps`` -- list of elements in the ground set of matroid M that are loops

    OUTPUT: a tuple containing 4 elements in this order:

    1. A dictionary containing 2-tuple (x,y) coordinates with
       ``M.simplify.groundset()`` elements that can be placed on the sides of
       the triangle as keys.
    2. A list of 3 lists of elements of ``M.simplify.groundset()`` that can
       be placed on the 3 sides of the triangle.
    3. A list of elements of ``M.simplify.groundset()`` that cane be placed
       inside the triangle in the geometric representation.
    4. A list of lists of elements of ``M.simplify.groundset()`` that
       correspond to lines in the geometric representation other than the sides
       of the triangle.

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers as mph
        sage: M = Matroid(ring=GF(2), matrix=[[1, 0, 0, 0, 1, 1, 1,0],
        ....: [0, 1, 0, 1, 0, 1, 1,0],[0, 0, 1, 1, 1, 0, 1,0]])
        sage: N = M.simplify()
        sage: B1 = list(N.basis())
        sage: nB1 = list(set(M.simplify().groundset())-set(B1))
        sage: pts,trilines,nontripts,curvedlines = mph.it(M,
        ....: B1,nB1,M.loops())
        sage: pts
        {1: (1.0, 0.0), 2: (1.5, 1.0), 3: (0.5, 1.0),
         4: (0, 0), 5: (1, 2), 6: (2, 0)}
        sage: trilines
        [[3, 4, 5], [2, 5, 6], [1, 4, 6]]
        sage: nontripts
        [0]
        sage: curvedlines
        [[0, 1, 5], [0, 2, 4], [0, 3, 6], [1, 2, 3], [1, 4, 6], [2, 5, 6],
         [3, 4, 5]]

    .. NOTE::

        This method does NOT do any checks.
    """
    tripts = [(0, 0), (1, 2), (2, 0)]
    pts = {}
    j = 0
    for i in B1:
        pts[i] = tripts[j]
        j = j + 1
    pairs = [[0, 1], [1, 2], [0, 2]]
    L1 = []
    L2 = []
    L3 = []
    for i in nB1:
        if M.is_dependent([i, B1[pairs[0][0]], B1[pairs[0][1]]]):
            # Add to L1
            L1.append(i)
        elif M.is_dependent([i, B1[pairs[1][0]], B1[pairs[1][1]]]):
            # Add to L2
            L2.append(i)
        elif M.is_dependent([i, B1[pairs[2][0]], B1[pairs[2][1]]]):
            # Add to L3
            L3.append(i)
    L = [L1, L2, L3]  # megalist
    lines = []  # the list of lines
    for i in range(1, len(L)+1):
        lines.append([B1[pairs[i-1][0]]])
        lines[i-1].extend(L[i-1])
        lines[i-1].extend([B1[pairs[i-1][1]]])
    # place triangle and L1,L2,L3
    for i in L:  # loop over megalist
        interval = 1/float(len(i)+1)
        pt1 = list(tripts[pairs[L.index(i)][0]])
        pt2 = list(tripts[pairs[L.index(i)][1]])
        for j in range(1, len(i)+1):
            # loop over L1,L2,L3
            cc = interval*j
            pts[i[j-1]] = (cc*pt1[0]+(1-cc)*pt2[0], cc*pt1[1]+(1-cc)*pt2[1])
    trilines = [set(x) for x in lines if len(x) >= 3]
    set_lps = set(lps)
    curvedlines = [list(sx.difference(set_lps))
                   for x in M.flats(2) if (sx := set(x)) not in trilines
                   and len(list(x)) >= 3]
    nontripts = [i for i in nB1 if i not in pts]
    trilines = [list(s) for s in trilines]
    return pts, trilines, nontripts, curvedlines


def trigrid(tripts) -> list[list]:
    """
    Return a grid of 4 points inside given 3 points as a list.

    INPUT:

    - ``tripts`` -- list of 3 lists of the form [x,y] where x and y are the
      Cartesian coordinates of a point

    OUTPUT: list of lists containing 4 points in following order:

    - 1. Barycenter of 3 input points.
    - 2,3,4. Barycenters of 1. with 3 different 2-subsets of input points
      respectively.

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: points = matroids_plot_helpers.trigrid([[2,1],[4,5],[5,2]])
        sage: points
        [[3.6666666666666665, 2.6666666666666665],
         [3.222222222222222, 2.888888888888889],
         [4.222222222222222, 3.222222222222222],
         [3.5555555555555554, 1.8888888888888886]]

    .. NOTE::

        This method does NOT do any checks.
    """
    pairs = [[0, 1], [1, 2], [0, 2]]
    cpt = [float(tripts[0][0] + tripts[1][0] + tripts[2][0]) / 3,
           float(tripts[0][1] + tripts[1][1] + tripts[2][1]) / 3]
    grid = [cpt]
    for p, q in pairs:
        pt = [float(tripts[p][0] + tripts[q][0] + cpt[0]) / 3,
              float(tripts[p][1] + tripts[q][1] + cpt[1]) / 3]
        grid.append(pt)
    return grid


def addnontripts(tripts_labels, nontripts_labels, ptsdict) -> dict:
    """
    Return modified ``ptsdict`` with additional keys and values corresponding
    to ``nontripts``.

    INPUT:

    - ``tripts`` -- list of 3 ground set elements that are to be placed on
      vertices of the triangle
    - ``ptsdict`` -- dictionary (at least) containing ground set elements in
      ``tripts`` as keys and their (x,y) position as values
    - ``nontripts`` -- list of ground set elements whose corresponding points
      are to be placed inside the triangle

    OUTPUT:

    A dictionary containing ground set elements in ``tripts`` as keys and
    their (x,y) position as values along with all keys and respective values
    in ``ptsdict``.

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: from sage.matroids.advanced import setprint
        sage: ptsdict={'a':(0,0),'b':(1,2),'c':(2,0)}
        sage: ptsdict_1=matroids_plot_helpers.addnontripts(['a','b','c'],
        ....:         ['d','e','f'],ptsdict)
        sage: setprint(ptsdict_1)
        {'a': [0, 0], 'b': [1, 2], 'c': [0, 2], 'd': [0.6666666666666666, 1.0],
        'e': [0.6666666666666666, 0.8888888888888888],
        'f': [0.8888888888888888, 1.3333333333333333]}
        sage: ptsdict_2=matroids_plot_helpers.addnontripts(['a','b','c'],
        ....:         ['d','e','f','g','h'],ptsdict)
        sage: setprint(ptsdict_2)
        {'a': [0, 0], 'b': [1, 2], 'c': [0, 2], 'd': [0.6666666666666666, 1.0],
        'e': [0.6666666666666666, 0.8888888888888888],
        'f': [0.8888888888888888, 1.3333333333333333],
        'g': [0.2222222222222222, 1.0],
        'h': [0.5185185185185185, 0.5555555555555555]}

    .. NOTE::

        This method does NOT do any checks.
    """
    tripts = [list(ptsdict[p]) for p in tripts_labels]
    pairs = [[0, 1], [1, 2], [0, 2]]
    q = [tripts]
    num = len(nontripts_labels)
    gridpts = [[float((tripts[0][0] + tripts[1][0] + tripts[2][0]) / 3),
               float(tripts[0][1] + tripts[1][1] + tripts[2][1]) / 3]]
    n = 0
    while n < num + 1:
        g = trigrid(q[0])
        q.extend([[g[0], q[0][pairs[0][0]], q[0][pairs[0][1]]],
                  [g[0], q[0][pairs[1][0]], q[0][pairs[1][1]]],
                  [g[0], q[0][pairs[2][0]], q[0][pairs[2][1]]]])
        q.remove(q[0])
        gridpts.extend(g[1:])
        if n == 0:
            n = n + 4
        else:
            n = n + 3
    for j, p in enumerate(nontripts_labels):
        ptsdict[p] = tuple(gridpts[j])
    return ptsdict


def createline(ptsdict, ll, lineorders2=None) -> tuple[list, list, list, list]:
    """
    Return ordered lists of coordinates of points to be traversed to draw a
    2D line.

    INPUT:

    - ``ptsdict`` -- dictionary containing keys and their (x,y) position as
      values
    - ``ll`` -- list of keys in ``ptsdict`` through which a line is to be
      drawn
    - ``lineorders2`` -- (optional) list of ordered lists of keys in
      ``ptsdict`` such that if ll is setwise same as any of these then points
      corresponding to values of the keys will be traversed in that order thus
      overriding internal order deciding heuristic

    OUTPUT: a tuple containing 4 elements in this order:

    1. Ordered list of x-coordinates of values of keys in ``ll`` specified in
       ptsdict.
    2. Ordered list of y-coordinates of values of keys in ``ll`` specified
       in ptsdict.
    3. Ordered list of interpolated x-coordinates of points through which a
       line can be drawn.
    4. Ordered list of interpolated y-coordinates of points through which a
       line can be drawn.

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: ptsdict = {'a':(1,3),'b':(2,1),'c':(4,5),'d':(5,2)}
        sage: x,y,x_i,y_i = matroids_plot_helpers.createline(ptsdict,
        ....: ['a','b','c','d'])
        sage: [len(x), len(y), len(x_i), len(y_i)]
        [4, 4, 100, 100]
        sage: G = line(zip(x_i, y_i), color='black', thickness=3, zorder=1)             # needs sage.plot
        sage: G += points(zip(x, y), color='black', size=300, zorder=2)                 # needs sage.plot
        sage: G.show()                                                                  # needs sage.plot
        sage: x,y,x_i,y_i = matroids_plot_helpers.createline(ptsdict,
        ....: ['a','b','c','d'],lineorders2=[['b','a','c','d'],
        ....: ['p','q','r','s']])
        sage: [len(x), len(y), len(x_i), len(y_i)]
        [4, 4, 100, 100]
        sage: G = line(zip(x_i, y_i), color='black', thickness=3, zorder=1)             # needs sage.plot
        sage: G += points(zip(x, y), color='black', size=300, zorder=2)                 # needs sage.plot
        sage: G.show()                                                                  # needs sage.plot

    .. NOTE::

        This method does NOT do any checks.
    """
    x, lo = line_hasorder(ll, lineorders2)
    flip = False
    if not x:
        # convert dictionary to list of lists
        linepts = [list(ptsdict[i]) for i in ll]
        xpts = [xx[0] for xx in linepts]
        ypts = [yy[1] for yy in linepts]
        xdim = (float(max(xpts))-float(min(xpts)))
        ydim = (float(max(ypts))-float(min(ypts)))
        if xdim > ydim:
            sortedind = sorted(range(len(xpts)), key=lambda k: float(xpts[k]))
        else:
            sortedind = sorted(range(len(ypts)), key=lambda k: float(ypts[k]))
            flip = True
        sortedlinepts = [linepts[i] for i in sortedind]
        sortedx = [k[0] for k in sortedlinepts]
        sortedy = [k[1] for k in sortedlinepts]
    else:
        linepts = [list(ptsdict[i]) for i in lo]
        sortedx = [k[0] for k in linepts]
        sortedy = [k[1] for k in linepts]

    if flip:
        tck, u = scipy.interpolate.splprep([sortedy, sortedx], s=0.0, k=2)
        y_i, x_i = scipy.interpolate.splev(np.linspace(0, 1, 100), tck)
    else:
        tck, u = scipy.interpolate.splprep([sortedx, sortedy], s=0.0, k=2)
        x_i, y_i = scipy.interpolate.splev(np.linspace(0, 1, 100), tck)
    return sortedx, sortedy, x_i, y_i


def slp(M1, pos_dict=None, B=None) -> tuple:
    """
    Return simple matroid, loops and parallel elements of given matroid.

    INPUT:

    - ``M1`` -- matroid
    - ``pos_dict`` -- (optional) dictionary containing non loopy elements of
      ``M`` as keys and their (x,y) positions as keys. While simplifying the
      matroid, all except one element in a parallel class that is also
      specified in ``pos_dict`` will be retained.
    - ``B`` -- (optional) a basis of M1 that has been chosen for placement on
      vertices of triangle

    OUTPUT: a tuple containing 3 elements in this order:

    1. Simple matroid corresponding to ``M1``.
    2. Loops of matroid ``M1``.
    3. Elements that are in `M1.groundset()` but not in ground set of 1 or
       in 2

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: from sage.matroids.advanced import setprint
        sage: M1 = Matroid(ring=GF(2), matrix=[[1, 0, 0, 0, 1, 1, 1,0,1,0,1],
        ....:                                  [0, 1, 0, 1, 0, 1, 1,0,0,1,0],
        ....:                                  [0, 0, 1, 1, 1, 0, 1,0,0,0,0]])
        sage: M, L, P = matroids_plot_helpers.slp(M1)                                   # needs sage.rings.finite_rings
        sage: M.is_simple()                                                             # needs sage.rings.finite_rings
        True
        sage: setprint([L,P])                                                           # needs sage.rings.finite_rings
        [{10, 8, 9}, {7}]
        sage: M1 = Matroid(ring=GF(2), matrix=[[1, 0, 0, 0, 1, 1, 1,0,1,0,1],
        ....:                                  [0, 1, 0, 1, 0, 1, 1,0,0,1,0],
        ....:                                  [0, 0, 1, 1, 1, 0, 1,0,0,0,0]])
        sage: posdict = {8: (0, 0),  1: (2, 0),  2: (1, 2),  3: (1.5, 1.0),
        ....:            4: (0.5, 1.0),  5: (1.0, 0.0), 6: (1.0, 0.6666666666666666)}
        sage: M, L, P = matroids_plot_helpers.slp(M1, pos_dict=posdict)                 # needs sage.rings.finite_rings
        sage: M.is_simple()                                                             # needs sage.rings.finite_rings
        True
        sage: setprint([L,P])                                                           # needs sage.rings.finite_rings
        [{0, 10, 9}, {7}]

    .. NOTE::

        This method does NOT do any checks.
    """
    L = set(M1.loops())
    nP = L | set(M1.simplify().groundset())
    P = set(M1.groundset()) - nP
    if P:
        if pos_dict is not None:
            pcls = list({frozenset(set(M1.closure([p])) - L)
                         for p in list(P)})
            newP = []
            for pcl in pcls:
                pcl_in_dict = [p for p in list(pcl) if p in pos_dict.keys()]
                newP.extend(list(pcl - set([pcl_in_dict[0]])))
            return [M1.delete(L | set(newP)), L, set(newP)]
        if B is not None:
            pcls = list({frozenset(set(M1.closure([p])) - L)
                         for p in list(P)})
            newP = []
            for pcl in pcls:
                pcl_list = list(pcl)
                pcl_in_basis = [p for p in pcl_list if p in B]
                if pcl_in_basis:
                    newP.extend(list(pcl - set([pcl_in_basis[0]])))
                else:
                    newP.extend(list(pcl - set([pcl_list[0]])))
            return [M1.delete(L | set(newP)), L, set(newP)]
        return [M1.delete(L | P), L, P]
    return [M1.delete(L | P), L, P]


def addlp(M, M1, L, P, ptsdict, G=None, limits=None) -> tuple:
    """
    Return a graphics object containing loops (in inset) and parallel elements
    of matroid.

    INPUT:

    - ``M`` -- matroid
    - ``M1`` -- a simple matroid corresponding to ``M``
    - ``L`` -- list of elements in ``M.groundset()`` that are loops of matroid
      ``M``
    - ``P`` -- list of elements in ``M.groundset()`` not in
      ``M.simplify.groundset()`` or ``L``
    - ``ptsdict`` -- dictionary containing elements in ``M.groundset()`` not
      necessarily containing elements of ``L``
    - ``G`` -- (optional) a sage graphics object to which loops and parallel
      elements of matroid `M` added
    - ``limits`` -- (optional) current axes limits [xmin,xmax,ymin,ymax]

    OUTPUT: a 2-tuple containing:

    1. A sage graphics object containing loops and parallel elements of
       matroid ``M``
    2. axes limits array

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: M = Matroid(ring=GF(2), matrix=[[1, 0, 0, 0, 1, 1, 1,0,1],
        ....:                                 [0, 1, 0, 1, 0, 1, 1,0,0],
        ....:                                 [0, 0, 1, 1, 1, 0, 1,0,0]])
        sage: M1, L, P = matroids_plot_helpers.slp(M)                                   # needs sage.rings.finite_rings
        sage: G, lims = matroids_plot_helpers.addlp(M,M1,L,P,{0:(0,0)})                 # needs sage.plot sage.rings.finite_rings
        sage: G.show(axes=False)                                                        # needs sage.plot sage.rings.finite_rings

    .. NOTE::

        This method does NOT do any checks.
    """
    if G is None:
        G = Graphics()
    # deal with loops
    if L:
        loops = L
        looptext = ", ".join(str(l) for l in loops)
        if limits is None:
            rectx = -1
            recty = -1
        else:
            rectx = limits[0]
            recty = limits[2]-1
        rectw = 0.5 + 0.4*len(loops) + 0.5  # controlled based on len(loops)
        recth = 0.6
        G += polygon2d([[rectx, recty], [rectx, recty+recth],
                        [rectx+rectw, recty+recth], [rectx+rectw, recty]],
                       color='black', fill=False, thickness=4)
        G += text(looptext, (rectx+0.5, recty+0.3), color='black',
                  fontsize=13)
        G += point((rectx+0.2, recty+0.3), color=Color('#BDBDBD'), size=300,
                   zorder=2)
        G += text('Loop(s)', (rectx+0.5+0.4*len(loops)+0.1, recty+0.3),
                  fontsize=13, color='black')
        limits = tracklims(limits, [rectx, rectx+rectw], [recty, recty+recth])
    # deal with parallel elements
    if P:
        # create list of lists where inner lists are parallel classes
        pcls = []
        gnd = sorted(M1.groundset())
        for g in gnd:
            pcl = [g]
            for p in P:
                if M.rank([g, p]) == 1:
                    pcl.extend([p])
            pcls.append(pcl)
        ext_gnd = list(M.groundset())
        for pcl in pcls:
            if len(pcl) > 1:
                basept = list(ptsdict[pcl[0]])
                if len(pcl) <= 2:
                    # add side by side
                    ptsdict[pcl[1]] = (basept[0], basept[1]-0.13)
                    G += points(zip([basept[0]], [basept[1]-0.13]),
                                color=Color('#BDBDBD'), size=300, zorder=2)
                    G += text(pcl[0], (float(basept[0]),
                              float(basept[1])), color='black',
                              fontsize=13)
                    G += text(pcl[1], (float(basept[0]),
                              float(basept[1])-0.13), color='black',
                              fontsize=13)
                    limits = tracklims(limits, [basept[0]], [basept[1]-0.13])
                else:
                    # add in a bracket
                    pce = sorted([str(kk) for kk in pcl])
                    l = newlabel(set(ext_gnd))
                    ext_gnd.append(l)
                    G += text(l+'={ '+", ".join(pce)+' }', (float(basept[0]),
                              float(basept[1]-0.2)-0.034), color='black',
                              fontsize=13)
                    G += text(l, (float(basept[0]),
                              float(basept[1])), color='black',
                              fontsize=13)
                    limits = tracklims(limits, [basept[0]],
                                       [(basept[1]-0.2)-0.034])
    return G, limits


def line_hasorder(l, lodrs=None) -> tuple[bool, list]:
    """
    Determine if an order is specified for a line.

    INPUT:

    - ``l`` -- a line specified as a list of ground set elements
    - ``lordrs`` -- (optional) list of lists each specifying an order on
      a subset of ground set elements that may or may not correspond to a
      line in geometric representation

    OUTPUT: a tuple containing 2 elements in this order:

    1. A boolean indicating whether there is any list in ``lordrs`` that is
       setwise equal to ``l``.
    2. A list specifying an order on ``set(l)`` if 1. is True, otherwise
       an empty list.

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: matroids_plot_helpers.line_hasorder(['a','b','c','d'],
        ....: [['a','c','d','b'],['p','q','r']])
        (True, ['a', 'c', 'd', 'b'])
        sage: matroids_plot_helpers.line_hasorder(['a','b','c','d'],
        ....: [['p','q','r'],['l','m','n','o']])
        (False, [])

    .. NOTE::

        This method does NOT do any checks.
    """
    if lodrs is not None:
        set_l = Set(l)
        if lodrs:
            for i in lodrs:
                if Set(i) == set_l:
                    return True, i
    return False, []


def lineorders_union(lineorders1, lineorders2) -> list:
    """
    Return a list of ordered lists of ground set elements that corresponds to
    union of two sets of ordered lists of ground set elements in a sense.

    INPUT:

    - ``lineorders1`` -- list of ordered lists specifying orders on subsets
      of ground set
    - ``lineorders2`` -- list of ordered lists specifying orders subsets of
      ground set

    OUTPUT:

    A list of ordered lists of ground set elements that are (setwise) in only
    one of ``lineorders1`` or ``lineorders2`` along with the ones in
    lineorder2 that are setwise equal to any list in lineorders1.

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: matroids_plot_helpers.lineorders_union([['a','b','c'],
        ....: ['p','q','r'],['i','j','k','l']],[['r','p','q']])
        [['a', 'b', 'c'], ['p', 'q', 'r'], ['i', 'j', 'k', 'l']]
    """
    if lineorders1 is not None and lineorders2 is not None:
        lineorders = lineorders1
        for order in lineorders2:
            x, lo = line_hasorder(order, lineorders1)
            if not x:
                lineorders.append(order)
                lineorders.remove(lo)
        return lineorders
    if lineorders1 is None and lineorders2 is not None:
        return lineorders2
    if lineorders1 is not None:
        return lineorders1
    return None


def posdict_is_sane(M1, pos_dict) -> bool:
    """
    Return a boolean establishing sanity of ``posdict`` wrt matroid ``M``.

    INPUT:

    - ``M1`` -- matroid
    - ``posdict`` -- dictionary mapping ground set elements to (x,y)
      positions

    OUTPUT:

    A boolean that is ``True`` if posdict indeed has all the required elements
    to plot the geometric elements, otherwise ``False``.

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: M1 = Matroid(ring=GF(2), matrix=[[1, 0, 0, 0, 1, 1, 1,0,1,0,1],
        ....:                                  [0, 1, 0, 1, 0, 1, 1,0,0,1,0],
        ....:                                  [0, 0, 1, 1, 1, 0, 1,0,0,0,0]])
        sage: pos_dict = {0: (0, 0),  1: (2, 0),  2: (1, 2),  3: (1.5, 1.0),
        ....: 4: (0.5, 1.0),  5: (1.0, 0.0), 6: (1.0, 0.6666666666666666)}
        sage: matroids_plot_helpers.posdict_is_sane(M1,pos_dict)                        # needs sage.rings.finite_rings
        True
        sage: pos_dict = {1: (2, 0),  2: (1, 2),  3: (1.5, 1.0),
        ....:             4: (0.5, 1.0), 5: (1.0, 0.0), 6: (1.0, 0.6666666666666666)}
        sage: matroids_plot_helpers.posdict_is_sane(M1,pos_dict)                        # needs sage.rings.finite_rings
        False

    .. NOTE::

        This method does NOT do any checks. ``M1`` is assumed to be a
        matroid and ``posdict`` is assumed to be a dictionary.
    """
    L = set(M1.loops())
    nP = L | set(M1.simplify().groundset())
    P = set(M1.groundset()) - nP
    pcls = list({frozenset(set(M1.closure([p])) - L) for p in list(P)})
    for pcl in pcls:
        if not any(x in pos_dict for x in pcl):
            return False
    allP = []
    for pcl in pcls:
        allP.extend(pcl)
    return all(x in pos_dict
               for x in list(set(M1.groundset()) - (L | set(allP))))


def tracklims(lims, x_i=[], y_i=[]) -> list:
    """
    Return modified limits list.

    INPUT:

    - ``lims`` -- list with 4 elements ``[xmin,xmax,ymin,ymax]``
    - ``x_i`` -- new x values to track
    - ``y_i`` -- new y values to track

    OUTPUT: list with 4 elements ``[xmin,xmax,ymin,ymax]``

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: matroids_plot_helpers.tracklims([0,5,-1,7],[1,2,3,6,-1],
        ....: [-1,2,3,6])
        [-1, 6, -1, 7]

    .. NOTE::

        This method does NOT do any checks.
    """
    if lims is not None and lims[0] is not None and lims[1] is not None and \
       lims[2] is not None and lims[3] is not None:
        lims = [min(*x_i, lims[0]), max(*x_i, lims[1]),
                min(*y_i, lims[2]), max(*y_i, lims[3])]
    else:
        lims = [min(x_i), max(x_i), min(y_i), max(y_i)]
    return lims


def geomrep(M1, B1=None, lineorders1=None, pd=None, sp=False):
    """
    Return a sage graphics object containing geometric representation of
    matroid M1.

    INPUT:

    - ``M1`` -- matroid
    - ``B1`` -- (optional) list of elements in ``M1.groundset()`` that
      correspond to a basis of ``M1`` and will be placed as vertices of the
      triangle in the geometric representation of ``M1``
    - ``lineorders1`` -- (optional) list of ordered lists of elements of
      ``M1.groundset()`` such that if a line in geometric representation is
      setwise same as any of these then points contained will be traversed in
      that order thus overriding internal order deciding heuristic
    - ``pd`` -- (optional) dictionary mapping ground set elements to their
      (x,y) positions
    - ``sp`` -- (optional) if ``True``, a positioning dictionary and line orders
      will be placed in ``M._cached_info``

    OUTPUT:

    A sage graphics object of type <class 'sage.plot.graphics.Graphics'> that
    corresponds to the geometric representation of the matroid.

    EXAMPLES::

        sage: from sage.matroids import matroids_plot_helpers
        sage: M = matroids.catalog.P7()
        sage: G = matroids_plot_helpers.geomrep(M)                                      # needs sage.plot
        sage: G.show(xmin=-2, xmax=3, ymin=-2, ymax=3)                                  # needs sage.plot
        sage: M = matroids.catalog.P7()
        sage: G = matroids_plot_helpers.geomrep(M, lineorders1=[['f','e','d']])         # needs sage.plot
        sage: G.show(xmin=-2, xmax=3, ymin=-2, ymax=3)                                  # needs sage.plot

    .. NOTE::

        This method does NOT do any checks.
    """
    G = Graphics()
    # create lists of loops and parallel elements and simplify given matroid
    M, L, P = slp(M1, pos_dict=pd, B=B1)
    if B1 is None:
        B1 = list(M.basis())
    M._cached_info = M1._cached_info

    if M.rank() == 0:
        limits = None
        loops = L
        looptext = ", ".join(str(l) for l in loops)
        rectx = -1
        recty = -1
        rectw = 0.5 + 0.4 * len(loops) + 0.5  # controlled based on len(loops)
        recth = 0.6
        G += polygon2d([[rectx, recty], [rectx, recty+recth],
                        [rectx+rectw, recty+recth], [rectx+rectw, recty]],
                       color='black', fill=False, thickness=4)
        G += text(looptext, (rectx+0.5, recty+0.3), color='black',
                  fontsize=13)
        G += point((rectx+0.2, recty+0.3), color=Color('#BDBDBD'), size=300,
                   zorder=2)
        G += text('Loop(s)', (rectx+0.5+0.4*len(loops)+0.1, recty+0.3),
                  fontsize=13, color='black')
        limits = tracklims(limits, [rectx, rectx+rectw], [recty, recty+recth])
        G.axes(False)
        G.axes_range(xmin=limits[0]-0.5, xmax=limits[1]+0.5,
                     ymin=limits[2]-0.5, ymax=limits[3]+0.5)
        return G
    if M.rank() == 1:
        if M._cached_info is not None and \
           'plot_positions' in M._cached_info.keys() and \
           M._cached_info['plot_positions'] is not None:
            pts = M._cached_info['plot_positions']
        else:
            pts = {}
            gnd = sorted(M.groundset())
        pts[gnd[0]] = (1, float(2)/3)
        G += point((1, float(2)/3), size=300, color=Color('#BDBDBD'), zorder=2)
        pt = [1, float(2)/3]
        if not P:
            G += text(gnd[0], (float(pt[0]), float(pt[1])), color='black',
                      fontsize=13)
        pts2 = pts
        # track limits [xmin,xmax,ymin,ymax]
        pl = [list(x) for x in pts2.values()]
        lims = tracklims([None, None, None, None], [pnt[0] for pnt in pl],
                         [pnt[1] for pnt in pl])
    elif M.rank() == 2:
        nB1 = set(M.groundset()) - set(B1)
        bline = [j for j in nB1 if M.is_dependent([j, B1[0], B1[1]])]
        interval = len(bline)+1
        if M._cached_info is not None and \
           'plot_positions' in M._cached_info.keys() and \
           M._cached_info['plot_positions'] is not None:
            pts2 = M._cached_info['plot_positions']
        else:
            pts2 = {}
            pts2[B1[0]] = (0, 0)
            pts2[B1[1]] = (2, 0)
            lpt = list(pts2[B1[0]])
            rpt = list(pts2[B1[1]])
            for k in range(len(bline)):
                cc = (float(1)/interval)*(k+1)
                pts2[bline[k]] = (cc*lpt[0]+(1-cc)*rpt[0],
                                  cc*lpt[1]+(1-cc)*rpt[1])
            if sp:
                M._cached_info['plot_positions'] = pts2
        # track limits [xmin,xmax,ymin,ymax]
        pl = [list(x) for x in pts2.values()]
        lims = tracklims([None, None, None, None], [pt[0] for pt in pl],
                         [pt[1] for pt in pl])
        bline.extend(B1)
        ptsx, ptsy, x_i, y_i = createline(pts2, bline, lineorders1)
        lims = tracklims(lims, x_i, y_i)
        G += line(zip(x_i, y_i), color='black', thickness=3, zorder=1)
        pels = [p for p in pts2
                if any(M1.rank([p, q]) == 1 for q in P)]
        allpts = [list(pts2[i]) for i in M.groundset()]
        xpts = [float(k[0]) for k in allpts]
        ypts = [float(k[1]) for k in allpts]
        G += points(zip(xpts, ypts), color=Color('#BDBDBD'), size=300,
                    zorder=2)
        for i in pts2:
            if i not in pels:
                pt = list(pts2[i])
                G += text(i, (float(pt[0]), float(pt[1])), color='black',
                          fontsize=13)
    else:
        if M._cached_info is None or \
           'plot_positions' not in M._cached_info.keys() or \
           M._cached_info['plot_positions'] is None:
            (pts, trilines,
             nontripts, curvedlines) = it(M1, B1,
                                          list(set(M.groundset())-set(B1)),
                                          list(set(L) | set(P)))
            pts2 = addnontripts([B1[0], B1[1], B1[2]], nontripts, pts)
            trilines.extend(curvedlines)
        else:
            pts2 = M._cached_info['plot_positions']
            trilines = [list(set(x).difference(L | P))
                        for x in M1.flats(2) if len(list(x)) >= 3]
        pl = [list(x) for x in pts2.values()]
        lims = tracklims([None, None, None, None], [pt[0] for pt in pl],
                         [pt[1] for pt in pl])
        for ll in trilines:
            if len(ll) >= 3:
                ptsx, ptsy, x_i, y_i = createline(pts2, ll, lineorders1)
                lims = tracklims(lims, x_i, y_i)
                G += line(zip(x_i, y_i), color='black', thickness=3, zorder=1)
        pels = [p for p in pts2
                if any(M1.rank([p, q]) == 1 for q in P)]
        allpts = [list(pts2[i]) for i in M.groundset()]
        xpts = [float(k[0]) for k in allpts]
        ypts = [float(k[1]) for k in allpts]
        G += points(zip(xpts, ypts), color=Color('#BDBDBD'), size=300,
                    zorder=2)
        for i in pts2:
            if i not in pels:
                pt = list(pts2[i])
                G += text(i, (float(pt[0]), float(pt[1])), color='black',
                          fontsize=13)
        if sp:
            M1._cached_info['plot_positions'] = pts2
            M1._cached_info['plot_lineorders'] = lineorders1
    # deal with loops and parallel elements
    G, lims = addlp(M1, M, L, P, pts2, G, lims)
    G.axes(False)
    G.axes_range(xmin=lims[0]-0.5, xmax=lims[1]+0.5, ymin=lims[2]-0.5,
                 ymax=lims[3]+0.5)
    return G



r"""
Families of Manifold Objects

The class :class:`ManifoldObjectFiniteFamily` is a subclass of :class:`FiniteFamily`
that provides an associative container of manifold objects, indexed by their
``_name`` attributes.

:class:`ManifoldObjectFiniteFamily` instances are totally ordered according
to their lexicographically ordered element names.

The subclass :class:`ManifoldSubsetFiniteFamily` customizes the print
representation further.

AUTHORS:

- Matthias Koeppe (2021): initial version
"""
# *****************************************************************************
#       Copyright (C) 2021 Matthias Koeppe <mkoeppe@math.ucdavis.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
# *****************************************************************************

from functools import total_ordering

from sage.sets.family import FiniteFamily


@total_ordering
class ManifoldObjectFiniteFamily(FiniteFamily):
    r"""
    Finite family of manifold objects, indexed by their names.

    The class :class:`ManifoldObjectFiniteFamily` inherits from
    :class:`FiniteFamily`.  Therefore it is an associative container.

    It provides specialized ``__repr__`` and ``_latex_`` methods.

    :class:`ManifoldObjectFiniteFamily` instances are totally ordered
    according to their lexicographically ordered element names.

    EXAMPLES::

        sage: from sage.manifolds.family import ManifoldObjectFiniteFamily
        sage: M = Manifold(2, 'M', structure='topological')
        sage: A = M.subset('A')
        sage: B = M.subset('B')
        sage: C = B.subset('C')
        sage: F = ManifoldObjectFiniteFamily([A, B, C]); F
        Set {A, B, C} of objects of the 2-dimensional topological manifold M
        sage: latex(F)
        \{A, B, C\}
        sage: F['B']
        Subset B of the 2-dimensional topological manifold M

    All objects must have the same base manifold::

        sage: N = Manifold(2, 'N', structure='topological')
        sage: ManifoldObjectFiniteFamily([M, N])
        Traceback (most recent call last):
        ...
        TypeError: all objects must have the same manifold
    """

    def __init__(self, objects=(), keys=None):
        r"""
        Initialize a new instance of :class:`ManifoldObjectFiniteFamily`.

        TESTS:

            sage: from sage.manifolds.family import ManifoldObjectFiniteFamily
            sage: M = Manifold(2, 'M', structure='topological')
            sage: A = M.subset('A')
            sage: B = M.subset('B')
            sage: C = B.subset('C')
            sage: F = ManifoldObjectFiniteFamily([A, B, C]); F
            Set {A, B, C} of objects of the 2-dimensional topological manifold M
            sage: TestSuite(F).run(skip='_test_elements')

        Like ``frozenset``, it can be created from any iterable::

            sage: from sage.manifolds.family import ManifoldSubsetFiniteFamily
            sage: M = Manifold(2, 'M', structure='topological')
            sage: I = M.subset('I')
            sage: gen = (subset for subset in (M, I, M, I, M, I)); gen
            <generator object ...>
            sage: ManifoldSubsetFiniteFamily(gen)
            Set {I, M} of subsets of the 2-dimensional topological manifold M
        """
        if isinstance(objects, dict):
            dictionary = objects
        else:
            dictionary = {object._name: object for object in objects}
        if keys is None:
            keys = sorted(dictionary.keys())
        FiniteFamily.__init__(self, dictionary, keys)
        names_and_latex_names = sorted(
            (object._name, object._latex_name) for object in self
        )
        self._name = '{' + ', '.join(keys) + '}'
        latex_names = (latex_name for name, latex_name in names_and_latex_names)
        self._latex_name = r'\{' + ', '.join(latex_names) + r'\}'
        try:
            object_iter = iter(self)
            self._manifold = next(object_iter)._manifold
        except StopIteration:
            self._manifold = None
        else:
            if not all(object._manifold == self._manifold for object in object_iter):
                raise TypeError(
                    f'all {self._repr_object_type()} must have the same manifold'
                )

    def _repr_object_type(self):
        r"""
        String that describes the type of the elements (plural).

        TESTS::

            sage: from sage.manifolds.family import ManifoldObjectFiniteFamily
            sage: M = Manifold(2, 'M', structure='topological')
            sage: A = M.subset('A')
            sage: B = M.subset('B')
            sage: ManifoldObjectFiniteFamily([A, B]).__repr__()           # indirect doctest
            'Set {A, B} of objects of the 2-dimensional topological manifold M'
        """
        return "objects"

    def __lt__(self, other):
        r"""
        Implement the total order on instances of :class:`ManifoldObjectFiniteFamily`.

        TESTS::

            sage: from sage.manifolds.family import ManifoldSubsetFiniteFamily
            sage: M = Manifold(2, 'M', structure='topological')
            sage: A = M.subset('A')
            sage: B = M.subset('B')
            sage: sorted([ManifoldSubsetFiniteFamily([A, B]), ManifoldSubsetFiniteFamily([]),
            ....:         ManifoldSubsetFiniteFamily([B]), ManifoldSubsetFiniteFamily([A])])
            [{},
             Set {A} of subsets of the 2-dimensional topological manifold M,
             Set {A, B} of subsets of the 2-dimensional topological manifold M,
             Set {B} of subsets of the 2-dimensional topological manifold M]
        """
        if not isinstance(other, ManifoldSubsetFiniteFamily):
            return NotImplemented
        return self.keys() < other.keys()

    def __repr__(self):
        r"""
        String representation of the object.

        TESTS::

            sage: from sage.manifolds.family import ManifoldObjectFiniteFamily
            sage: ManifoldObjectFiniteFamily().__repr__()
            '{}'
            sage: M = Manifold(2, 'M', structure='topological')
            sage: A = M.subset('A')
            sage: B = M.subset('B')
            sage: ManifoldObjectFiniteFamily([A, B]).__repr__()
            'Set {A, B} of objects of the 2-dimensional topological manifold M'
        """
        if self:
            return "Set {} of {} of the {}".format(
                self._name, self._repr_object_type(), self._manifold
            )
        return "{}"

    def _latex_(self):
        r"""
        LaTeX representation of ``self``.

        TESTS::

            sage: from sage.manifolds.family import ManifoldSubsetFiniteFamily
            sage: M = Manifold(2, 'M', structure='topological')
            sage: A = M.subset('A')
            sage: B = M.subset('B')
            sage: ManifoldSubsetFiniteFamily([B, A])._latex_()
            '\\{A, B\\}'
        """
        return self._latex_name


class ManifoldSubsetFiniteFamily(ManifoldObjectFiniteFamily):
    r"""
    Finite family of subsets of a topological manifold, indexed by their names.

    The class :class:`ManifoldSubsetFiniteFamily` inherits from
    :class:`ManifoldObjectFiniteFamily`.  It provides an associative
    container with specialized ``__repr__`` and ``_latex_`` methods.

    :class:`ManifoldSubsetFiniteFamily` instances are totally ordered according
    to their lexicographically ordered element (subset) names.

    EXAMPLES::

        sage: from sage.manifolds.family import ManifoldSubsetFiniteFamily
        sage: M = Manifold(2, 'M', structure='topological')
        sage: A = M.subset('A')
        sage: B = M.subset('B')
        sage: C = B.subset('C')
        sage: ManifoldSubsetFiniteFamily([A, B, C])
        Set {A, B, C} of subsets of the 2-dimensional topological manifold M
        sage: latex(_)
        \{A, B, C\}

    All subsets must have the same base manifold::

        sage: N = Manifold(2, 'N', structure='topological')
        sage: ManifoldSubsetFiniteFamily([M, N])
        Traceback (most recent call last):
        ...
        TypeError: all open subsets must have the same manifold
    """

    @classmethod
    def from_subsets_or_families(cls, *subsets_or_families):
        r"""
        Construct a ``ManifoldSubsetFiniteFamily`` from given subsets or
        iterables of subsets.

        EXAMPLES::

            sage: from sage.manifolds.family import ManifoldSubsetFiniteFamily
            sage: M = Manifold(2, 'M', structure='topological')
            sage: A = M.subset('A')
            sage: Bs = (M.subset(f'B{i}') for i in range(5))
            sage: Cs = ManifoldSubsetFiniteFamily([M.subset('C0'), M.subset('C1')])
            sage: ManifoldSubsetFiniteFamily.from_subsets_or_families(A, Bs, Cs)
            Set {A, B0, B1, B2, B3, B4, C0, C1} of subsets of the 2-dimensional topological manifold M
        """

        def generate_subsets():
            from sage.manifolds.subset import ManifoldSubset

            for arg in subsets_or_families:
                if isinstance(arg, ManifoldSubset):
                    yield arg
                else:
                    # arg must be an iterable of ManifoldSubset instances
                    yield from arg

        return cls(generate_subsets())

    def _repr_object_type(self):
        r"""
        String that describes the type of the elements (plural).

        TESTS::

            sage: from sage.manifolds.family import ManifoldSubsetFiniteFamily
            sage: M = Manifold(2, 'M', structure='topological')
            sage: A = M.subset('A')
            sage: B = M.subset('B')
            sage: ManifoldSubsetFiniteFamily([A, B]).__repr__()           # indirect doctest
            'Set {A, B} of subsets of the 2-dimensional topological manifold M'
        """
        if all(subset.is_open() for subset in self):
            return "open subsets"
        return "subsets"
